# -*- coding: utf-8 -*-
#  privacyIDEA is a fork of LinOTP
#  * May 08, 2014 Cornelius Kölbel
#  * Sep 17, 2014 Cornelius Kölbel, cornelius@privacyidea.org
#                 Improve the return value of the InitDetail
#  License:  AGPLv3
#  contact:  http://www.privacyidea.org
#
#  2014-10-03 Move the QR stuff in getInitDetail into the token classes
#             Cornelius Kölbel <cornelius@privacyidea.org>
#
#  Copyright (C) 2010 - 2014 LSE Leading Security Experts GmbH
#  License:  AGPLv3
#  contact:  http://www.linotp.org
#            http://www.lsexperts.de
#            linotp@lsexperts.de
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
'''

  Description:  This file containes the standard token definitions:
              - OCRATokenClass

              It also contains the base class "TokenClass", that you may use to
              define your own tokenclasses.

              You can add your own Tokens by adding the modules comma seperated to the directive
                  privacyideaTokenModules
              in the privacyidea.ini file.

  Dependencies: depends on several modules from privacyidea.lib but also in case of VascoTokenClass on
              privacyideaee.lib.ImportOTP.vasco


'''
import re
import binascii

import logging
import time
import hashlib
import datetime

import traceback


from privacyidea.lib.error import TokenAdminError
from privacyidea.lib.error import ParameterError

from privacyidea.lib.util import getParam
from privacyidea.lib.util import generate_otpkey
from privacyidea.lib.log import log_with

from privacyidea.lib.config  import getFromConfig

from privacyidea.lib.user import getUserResolverId
from privacyidea.lib.config import get_privacyIDEA_config
from privacyidea.lib.crypto import decryptPin
from privacyidea.lib.crypto import encryptPin
from privacyidea.lib.crypto import kdf2
from privacyidea.lib.crypto import urandom
from privacyidea.lib.crypto import createNonce

### TODO: move this as ocra specific methods
from privacyidea.lib.token import getRolloutToken4User
from privacyidea.lib.util import normalize_activation_code

from privacyidea.lib.ocra    import OcraSuite
from privacyidea.model       import OcraChallenge

from privacyidea.model.meta  import Session

from privacyidea.lib.validate import check_pin
from privacyidea.lib.validate import check_otp
from privacyidea.lib.validate import split_pin_otp
from privacyidea.lib.reply   import create_img


from sqlalchemy         import asc, desc
#from sqlalchemy.sql.expression import in_

from pylons.i18n.translation import _



# needed for ocra token
import urllib

import json


optional = True
required = False

log = logging.getLogger(__name__)

class TokenClass(object):

    @log_with(log)
    def __init__(self, token):
        self.type = ''
        self.token = token
        ## the info is a generic container, to store token specific processing info
        ## which could be retrieved in the controllers
        self.info = {}
        self.hKeyRequired = False
        self.mode = ['authenticate', 'challenge']

    def setType(self, typ):
        typ = u'' + typ
        self.type = typ
        self.token.setType(typ)

    @classmethod
    def getClassType(cls):
        return None

    @classmethod
    def getClassPrefix(cls):
        return "UNK"

    def getType(self):
        return self.token.getType()

    def addToInfo(self, key, value):
        self.info[key] = value
        return self.info

    def setInfo(self, info):
        if type(info) not in (dict):
            raise Exception("Info setting: wron data type - msut be dict")
        self.info = info
        return self.info

    @log_with(log)
    def getInfo(self):
        '''
        getInfo - return the status of the token rollout

        :return: return the status dict.
        :rtype: dict
        '''
        return self.info

    def checkOtp(self, anOtpVal1, counter, window, options=None):
        '''
        This checks the OTP value, AFTER the upper level did
        the checkPIN

        return:
            counter of the matching OTP value.
        '''
        return -1

    def getOtp(self, curtTime=""):
        '''
        The default token does not support getting the otp value
        will return something like:
            1, pin, otpval, combined

        a negative value is a failure.
        '''
        return (-2, 0, 0, 0)

    def get_multi_otp(self, count=0, epoch_start=0, epoch_end=0, curTime=None, timestamp=None):
        '''
        This returns a dictionary of multiple future OTP values of a token.

        parameter
        :param count: how many otp values should be returned
        :param epoch_start: time based tokens: start when
        :param epoch_end: time based tokens: stop when
        :param curTime: current time for TOTP token (for selftest)
        :type curTime: datetime object
        :param timestamp: unix time, current time for TOTP token (for selftest)
        :type timestamp: int

        :return: True/False, error text, OTP dictionary
        :rtype: Tuple
        '''
        return (False, "get_multi_otp not implemented for this tokentype", {})

### new highlevel interface which covers the checkPin and checkOTP
    def authenticate(self, passw, user, options=None):
        '''
        This is the method that verifies single shot authentication like
        they are done with push button tokens.

        It is a high level interface to support as well other tokens, which
        do not have a pin and otp seperation - they could overwrite
        this method

        **remarks:** we have to call the global methods (check_pin,++) as they
        take the pin policies into account

        :param passw: the passw which could be pin+otp
        :type passw: string
        :param user: The authenticating user
        :type user: User object
        :param options: dictionary of additional request parameters
        :type options: (dict)

        :return: returns tuple true or false for the pin match, the otpcounter (int)
        and the reply (dict) that will be added as additional information in
        the JSON response of ``/validate/check``.
        '''

        pin_match = False
        otp_counter = -1
        reply = None

        (res, pin, otpval) = split_pin_otp(self, passw, user, options=options)
        if res != -1:
            pin_match = check_pin(self, pin, user=user, options=options)
            if pin_match is True:
                otp_counter = check_otp(self, otpval, options=options)

        return (pin_match, otp_counter, reply)


### challenge interfaces starts here
    def is_challenge_request(self, passw, user, options=None):
        '''
        This method checks, if this is a request, that triggers a challenge.

        The default behaviour to trigger a challenge is,
        if the ``passw`` parameter only contains the correct token pin *and*
        the request contains a ``data`` or a ``challenge`` key i.e. if the
        ``options`` parameter contains a key ``data`` or ``challenge``.

        Each token type can decide on its own under which condition a challenge
        is triggered by overwriting this method.

        **please note**: in case of pin policy == 2 (no pin is required)
        the ``check_pin`` would always return true! Thus each request
        containing a ``data`` or ``challenge`` would trigger a challenge!

        :param passw: password, which might be pin or pin+otp
        :type passw: string
        :param user: The user from the authentication request
        :type user: User object
        :param options: dictionary of additional request parameters
        :type options: dict

        :return: true or false
        '''

        request_is_valid = False

        pin_match = check_pin(self, passw, user=user, options=options)
        if pin_match is True:
            if "data" in options or "challenge" in options:
                request_is_valid = True

        return request_is_valid

    def is_challenge_response(self, passw, user, options=None, challenges=None):
        '''
        This method checks, if this is a request, that is the response to
        a previously sent challenge.

        The default behaviour to check if this is the response to a
        previous challenge is simply by checking if the request contains
        a parameter ``state`` or ``transactionid`` i.e. checking if the
        ``options`` parameter contains a key ``state`` or ``transactionid``.

        This method does not try to verify the response itself!
        It only determines, if this is a response for a challenge or not.

        :param passw: password, which might be pin or pin+otp
        :type passw: string
        :param user: the requesting user
        :type user: User object
        :param options: dictionary of additional request parameters
        :type options: (dict)
        :param challenges: A list of challenges for this token. These challenges may be used, to identify
        if this request is a response for a challenge.

        :return: true or false
        '''

        challenge_response = False
        if "state" in options or "transactionid" in options:
            challenge_response = True

        ## we leave out the checkOtp, which is done later
        ## either in checkResponse4Challenge
        ## or in the check pin+otp

        return challenge_response


    def is_challenge_valid(self, challenge=None):
        '''
        This method verifies if the given challenge is still valid.

        The default implementation checks, if the challenge start is in the
        default validity time window.

        **Please note**: This method does not check the response for the
        challenge itself. This is done by the method
        :py:meth:`~privacyidea.lib.tokenclass.TokenClass.checkResponse4Challenge`.
        E.g. this very method ``is_challenge_valid`` is used by the method
        :py:meth:`~privacyidea.lib.tokenclass.TokenClass.challenge_janitor`
        to clean up old challenges.

        :param challenge: The challenge to be checked
        :type challenge: challenge object
        :return: true or false
        '''

        validity = 120
        ret = False

        try:
            validity = int(getFromConfig('DefaultChallengeValidityTime', 120))

            ## handle the token specific validity
            typ = self.getType()
            if typ == 'sms':
                lookup_for = 'SMSProviderTimeout'
            else:
                lookup_for = typ.capitalize() + 'ChallengeValidityTime'
            validity = int(getFromConfig(lookup_for, validity))

            ## instance specific timeout
            validity = int(self.getFromTokenInfo('challenge_validity_time',
                                                 validity))

        except ValueError:
            validity = 120

        if challenge is not None:
            c_start_time = challenge.get('timestamp')
            c_now = datetime.datetime.now()
            if (c_now < c_start_time + datetime.timedelta(seconds=validity)
                and c_now > c_start_time):
                ret = True

        return ret

    def initChallenge(self, transactionid, challenges=None, options=None):
        """
        This method initializes the challenge.

        This is a hook that is called before the method
        :py:meth:`~privacyidea.lib.tokenclass.TokenClass.createChallenge`, which
        will only be called if this method returns success==true.

        Thus this method can be used, to verify if there is an outstanding challenge
        or if a new challenge needs to be created.
        E.g. this hook can be used, to implement a blocking mechanism to
        allow the creation of a new challenge only after a certain timeout.
        If there is an already outstanding challenge the return value can refer to this.
        (s. ticket #2986)

        :param transactionid: the id of the new challenge
        :type transactionid: string
        :param options: the request parameters
        :type options: dict
        :param challenges: a list of all valid challenges for this token.
        :type challenges: list

        :return: tuple of ( success, transid, message, additional attributes )

        The ``transid`` (the best transaction id for this request context),
        ``message``, and additional ``attributes`` (dictionar) are displayed as results
        in the JSON response of the ``/validate/check`` request.

        Only in case of ``success`` == true the next method ``createChallenge`` will be called.
        """
        return (True, transactionid, 'challenge init ok', {})


    def checkResponse4Challenge(self, user, passw, options=None, challenges=None):
        '''
        This method verifies if the given ``passw`` matches any existing ``challenge``
        of the token.

        It then returns the new otp_counter of the token and the
        list of the matching challenges.

        In case of success the otp_counter needs to be > 0.
        The matching_challenges is passed to the method
        :py:meth:`~privacyidea.lib.tokenclass.TokenClass.challenge_janitor`
        to clean up challenges.

        :param user: the requesting user
        :type user: User object
        :param passw: the password (pin+otp)
        :type passw: string
        :param options:  additional arguments from the request, which could be token specific
        :type options: dict
        :param challenges: A sorted list of valid challenges for this token.
        :type challenges: list
        :return: tuple of (otpcounter and the list of matching challenges)

        '''
        otp_counter = -1
        transid = None
        matching = None
        matching_challenges = []

        if 'transactionid' in options or 'state' in options:
            ## fetch the transactionid
            transid = options.get('transactionid', None)
            if transid == None:
                transid = options.get('state', None)

        ## check if the transactionid is in the list of challenges
        if transid is not None:
            for challenge in challenges:
                if challenge.getTransactionId() == transid:
                    matching = challenge
                    break
            if matching is not None:
                otp_counter = check_otp(self, passw, options=options)
                if otp_counter >= 0:
                    matching_challenges.append(matching)

        return (otp_counter, matching_challenges)



    def challenge_janitor(self, matching_challenges, challenges):
        '''
        This is the default janitor for the challenges of a token.

        The idea is to delete all challenges, which have an id lower than
        the matching one. Other janitors could be implemented on a token base
        and overwrite this behaviour.

        **Remarks**: In later versions this will be the place to hook a dynamically
        loaded default token specific janitor.

        :param matching_challenges: the last matching challenge
        :type matching_challenges: list
        :param challenges: all current challenges
        :type challenges: list

        :return: list of all challenges, which should be deleted
        '''
        to_be_deleted = []
        if matching_challenges is not None and len(matching_challenges) > 0:
            match_id = 0
            for match in matching_challenges:
                match_id = max([match_id, int(match.get('id'))])

            for ch in challenges:
                if int(ch.get('id')) <= match_id:
                    to_be_deleted.append(ch)

        ## as well append all out dated challenges
        for ch in challenges:
            if self.is_challenge_valid(ch) is False:
                to_be_deleted.append(ch)

        return to_be_deleted


    def createChallenge(self, transactionid, options=None):
        '''
        This method creates a challenge, which is submitted to the user.
        The submitted challenge will be preserved in the challenge
        database.

        This method is called *after* the method
        :py:meth:`~privacyidea.lib.tokenclass.TokenClass.initChallenge`.

        :param transactionid: the id of this challenge
        :param options: the request context parameters / data
        :type options: dict
        :return: tuple of (bool, message, data, attributes)

        The return tuple builds up like this:

        ``bool`` if submit was successfull;
        ``message`` which is displayed in the JSON response;
        ``data`` is preserved in the challenge;
        additional ``attributes``, which are displayed in the JSON response.

        '''
        message = 'Otp: '
        data = {'serial' : self.token.getSerial()}
        attributes = None
        return (True, message, data, attributes)

    def flush(self):
        self.token.storeToken()
        Session.flush()
        Session.commit()
        return



    def update(self, param, reset_failcount=True):

        tdesc = getParam(param, "description", optional)
        if tdesc is not None:
            self.token.setDescription(tdesc)

        ## key_size as parameter overrules a prevoiusly set
        ## value e.g. in hashlib in the upper classes
        key_size = getParam(param, "keysize", optional)
        if key_size is None:
            key_size = 20

        ##
        ## process the otpkey:
        ##   if otpkey given - take this
        ##   if not given
        ##       if genkey == 1 : create one
        ##   if required and otpkey == None:
        ##      raise param Exception, that we require an otpkey
        ##
        otpKey = getParam(param, "otpkey", optional)
        genkey = int(getParam(param, "genkey", optional) or 0)

        assert (genkey in [0, 1]), "TokenClass supports only genkey in range [0,1] : %r" % genkey

        if genkey == 1 and otpKey is not None:
            raise ParameterError('[ParameterError] You may either specify genkey or otpkey, but not both!', id=344)

        if otpKey is not None:
            self.token.setHKey(otpKey, reset_failcount=reset_failcount)
        else:
            if genkey == 1:
                otpKey = self._genOtpKey_()

        ## otpKey still None?? - raise the exception
        if otpKey is None:
            if self.hKeyRequired == True:
                otpKey = getParam(param, "otpkey", required)

        if otpKey is not None:
            self.addToInfo('otpkey', otpKey)
            self.setOtpKey(otpKey)

        pin = getParam(param, "pin", optional)
        if pin is not None:
            storeHashed = True
            enc = getParam(param, "encryptpin", optional)
            if enc is not None and "true" == enc.lower():
                storeHashed = False
            self.token.setPin(pin, storeHashed)

        otplen = getParam(param, 'otplen', optional)
        if otplen is not None:
            self.setOtpLen(otplen)

        return

    def _genOtpKey_(self, otpkeylen=None):
        '''
        private method, to create an otpkey
        '''
        if otpkeylen is None:
            if hasattr(self, 'otpkeylen'):
                otpkeylen = getattr(self, 'otpkeylen')
            else:
                otpkeylen = 20
        return generate_otpkey(otpkeylen)

    def setDescription(self, description):
        self.token.setDescription(u'' + description)
        return

    def setDefaults(self):
        ## set the defaults

        self.token.privacyIDEAOtpLen = int(getFromConfig("DefaultOtpLen") or 6)
        self.token.privacyIDEACountWindow = int(getFromConfig("DefaultCountWindow") or 10)
        self.token.privacyIDEAMaxFail = int(getFromConfig("DefaultMaxFailCount") or 10)
        self.token.privacyIDEASyncWindow = int(getFromConfig("DefaultSyncWindow")or 1000)

        self.token.privacyIDEATokenType = u'' + self.type
        return

    def setUser(self, user, report):
        '''
        :param user: a User() object, consisting of loginname and realm
        :param report: tbdf.
        '''
        (uuserid, uidResolver, uidResolverClass) = getUserResolverId(user, report)
        self.token.privacyIDEAIdResolver = uidResolver
        self.token.privacyIDEAIdResClass = uidResolverClass
        self.token.privacyIDEAUserid = uuserid

    def getUser(self):
        uidResolver = self.token.privacyIDEAIdResolver
        uidResolverClass = self.token.privacyIDEAIdResClass
        uuserid = self.token.privacyIDEAUserid
        return (uuserid, uidResolver, uidResolverClass)

    def setUid(self, uid, uidResolver, uidResClass):
        '''
        sets the UID values in the database
        '''
        self.token.privacyIDEAIdResolver = uidResolver
        self.token.privacyIDEAIdResClass = uidResClass
        self.token.privacyIDEAUserid = uid
        return

    def reset(self):
        self.token.privacyIDEAFailCount = 0

    def addToSession(self, Session):
        Session.add(self.token)

    def deleteToken(self):
        self.token.deleteToken()

    def storeToken(self):
        self.token.storeToken()

    def resync(self, otp1, otp2, options=None):
        pass

    def getOtpCountWindow(self):
        return self.token.privacyIDEACountWindow

    def getOtpCount(self):
        return self.token.privacyIDEACount

    def isActive(self):
        return self.token.privacyIDEAIsactive

    def getFailCount(self):
        return self.token.privacyIDEAFailCount

    def setFailCount(self, failCount):
        self.token.privacyIDEAFailCount = failCount

    def getMaxFailCount(self):
        return self.token.privacyIDEAMaxFail

    def getUserId(self):
        return self.token.privacyIDEAUserid

    def setRealms(self, realms):
        self.token.setRealms(realms)

    def getSerial(self):
        return self.token.getSerial()
    
    def getTokentype(self):
        return self.token.getType()

    def setSoPin(self, soPin):
        self.token.setSoPin(soPin)

    def setUserPin(self, userPin):
        self.token.setUserPin(userPin)

    def setOtpKey(self, otpKey):
        self.token.setHKey(otpKey)

    def setOtpLen(self, otplen):
        self.token.privacyIDEAOtpLen = int(otplen)

    def getOtpLen(self):
        return self.token.privacyIDEAOtpLen

    def setOtpCount(self, otpCount):
        self.token.privacyIDEACount = int(otpCount)

    def setPin(self, pin, param=None):
        '''
        set the PIN. The optional parameter "param" can hold the information,
        if the PIN is encrypted or hashed.
        '''
        if param is None:
            param = {}
        storeHashed = True
        enc = getParam(param, "encryptpin", optional)
        if enc is not None and "true" == enc.lower():
            storeHashed = False

        self.token.setPin(pin, storeHashed)

    def getPinHashSeed(self):
        return self.token.privacyIDEAPinHash, self.token.privacyIDEASeed

    def setPinHashSeed(self, pinhash, seed):
        self.token.privacyIDEAPinHash = pinhash
        self.token.privacyIDEASeed = seed

    def enable(self, enable):
        self.token.privacyIDEAIsactive = enable

    def setMaxFail(self, maxFail):
        self.token.privacyIDEAMaxFail = maxFail

    def setHashLib(self, hashlib):
        self.addToTokenInfo("hashlib", hashlib)

    def incOtpFailCounter(self):
        log.debug('incOtpFailCounter')

        if self.token.privacyIDEAFailCount < self.token.privacyIDEAMaxFail:
            self.token.privacyIDEAFailCount = self.token.privacyIDEAFailCount + 1

        try:
            self.token.storeToken()
        except:
            log.error('update failed')
            raise TokenAdminError("Token Fail Counter update failed", id=1106)

        return self.token.privacyIDEAFailCount

    def setCounterWindow(self, countWindow):
        self.token.privacyIDEACountWindow = int(countWindow)

    def getCounterWindow(self):
        return self.token.privacyIDEACountWindow

    def setSyncWindow(self, syncWindow):
        self.token.privacyIDEASyncWindow = int(syncWindow)

    def getSyncWindow(self):
        return self.token.privacyIDEASyncWindow

    # hashlib algorithms:
    # http://www.doughellmann.com/PyMOTW/hashlib/index.html#module-hashlib

    def getHashlib(self, hLibStr):

        if hLibStr is None:
            return hashlib.sha1

        hashlibStr = hLibStr.lower()

        if hashlibStr == "md5":
            return hashlib.md5
        elif hashlibStr == "sha1":
            return hashlib.sha1
        elif hashlibStr == "sha224":
            return hashlib.sha224
        elif hashlibStr == "sha256":
            return hashlib.sha256
        elif hashlibStr == "sha384":
            return hashlib.sha384
        elif hashlibStr == "sha512":
            return hashlib.sha512
        else:
            return hashlib.sha1


    def getTokenInfo(self):
        info = {}

        tokeninfo = self.token.getInfo()
        try:
            info = json.loads(tokeninfo)
        except Exception as e:
            log.error('getTokenInfo %r' % (e))

        return info

    def setTokenInfo(self, info):

        if info is not None:
            tokeninfo = u'' + json.dumps(info, indent=0)
            self.token.setInfo(tokeninfo)


    def addToTokenInfo(self, key, value):
        info = {}
        tokeninfo = self.token.getInfo()

        if tokeninfo is not None:
            if len(tokeninfo) > 0 :
                info = json.loads(tokeninfo)

        #if info.has_key(key) == True:
        info[key] = value

        self.setTokenInfo(info)

    def getFromTokenInfo(self, key, default=None):
        ret = default

        info = self.getTokenInfo()

        if key in info:
            ret = info.get(key)
        return ret

    def set_count_auth_success_max(self, count):
        '''
        Sets the counter for the maximum allowed successful logins
        '''
        self.addToTokenInfo("count_auth_success_max", int(count))

    def set_count_auth_success(self, count):
        '''
        Sets the counter for the occurred successful logins
        '''
        self.addToTokenInfo("count_auth_success", int(count))

    def set_count_auth_max(self, count):
        '''
        Sets the counter for the maximum allowed login attemps
        '''
        self.addToTokenInfo("count_auth_max", int(count))

    def set_count_auth(self, count):
        '''
        Sets the counter for the occurred login attepms
        '''
        self.addToTokenInfo("count_auth", int(count))

    def get_count_auth_success_max(self):
        ret = 0
        try:
            ret = int(self.getFromTokenInfo("count_auth_success_max"))
        except:
            pass
        return ret

    def get_count_auth_success(self):
        ret = 0
        try:
            ret = int(self.getFromTokenInfo("count_auth_success"))
        except:
            pass
        return ret

    def get_count_auth_max(self):
        ret = 0
        try:
            ret = int(self.getFromTokenInfo("count_auth_max"))
        except:
            pass
        return ret

    def get_count_auth(self):
        ret = 0
        try:
            ret = int(self.getFromTokenInfo("count_auth"))
        except:
            pass
        return ret

    def get_validity_period_end(self):
        '''
        returns the end of validity period (if set)
        '''
        ret = ""
        try:
            ret = self.getFromTokenInfo("validity_period_end")
        except:
            pass
        return ret

    def set_validity_period_end(self, end_date):
        '''
        sets the end date of the validity period for a token
        '''
        ## upper layer will catch. we just try to verify the date format
        datetime.datetime.strptime(end_date, "%d/%m/%y %H:%M")

        self.addToTokenInfo("validity_period_end", end_date)

    def get_validity_period_start(self):
        '''
        returns the start of validity period (if set)
        '''
        ret = ""
        try:
            ret = self.getFromTokenInfo("validity_period_start")
        except:
            pass
        return ret

    def set_validity_period_start(self, start_date):
        '''
        sets the start date of the validity period for a token
        '''
        ##  upper layer will catch. we just try to verify the date format
        datetime.datetime.strptime(start_date, "%d/%m/%y %H:%M")
        self.addToTokenInfo("validity_period_start", start_date)


    def inc_count_auth_success(self):
        count = self.get_count_auth_success()
        count += 1
        self.set_count_auth_success(count)
        return count

    def inc_count_auth(self):
        count = self.get_count_auth()
        count += 1
        self.set_count_auth(count)
        return count

    def check_auth_counter(self):
        '''
        This function checks the count_auth and the count_auth_success
        '''
        if 0 != self.get_count_auth_max():
            if self.get_count_auth() >= self.get_count_auth_max():
                return False

        if 0 != self.get_count_auth_success_max():
            if self.get_count_auth_success() >= self.get_count_auth_success_max():
                return False

        return True

    def check_validity_period(self):
        '''
        This checks if the datetime.datetime.now() is within the validity period of the token.

        Returns either True/False
        '''
        start = self.get_validity_period_start()
        end = self.get_validity_period_end()

        check_start = False
        check_end = False
        try:
            dt_start = datetime.datetime.strptime(start, "%d/%m/%y %H:%M")
            check_start = True
        except:
            pass

        try:
            dt_end = datetime.datetime.strptime(end, "%d/%m/%y %H:%M")
            check_end = True
        except:
            pass

        if check_end:
            if dt_end < datetime.datetime.now():
                return False

        if check_start:
            if dt_start > datetime.datetime.now():
                return False

        return True

    @log_with(log)
    def incOtpCounter(self, counter, reset=True):
        '''
        method
            incOtpCounter(aToken, counter)

        parameters:
            token - a token object
            counter - the new counter
            reset - optional -

        exception:
            in case of an transaction fail an exception is thrown

        side effects:
            default of reset will reset the failCounter

        '''
        log.debug('incOtpCounter')

        resetCounter = False
        self.token.privacyIDEACount = counter + 1

        if reset == True:
            if getFromConfig("DefaultResetFailCount") == "True" :
                resetCounter = True

        #log.error("Serial %s: privacyIDEAFailCount %d < privacyIDEAMaxFail %d",token.privacyIDEATokenSerialnumber, token.privacyIDEAFailCount,token.privacyIDEAMaxFail)

        if resetCounter == True:
            if self.token.privacyIDEAFailCount < self.token.privacyIDEAMaxFail and self.token.privacyIDEAIsactive == True:
                self.token.privacyIDEAFailCount = 0;

        try:
            self.token.storeToken()

        except Exception as ex :
            log.error("Token Counter update failed: %r" % (ex))
            raise TokenAdminError("Token Counter update failed: %r" % (ex), id=1106)

        return self.token.privacyIDEACount


    def check_otp_exist(self, otp, window=None):
        '''
        checks if the given OTP value is/are values of this very token.
        This is used to autoassign and to determine the serial number of
        a token.
        '''
        return -1

    def splitPinPass(self, passw):

        res = 0
        try:
            otplen = int(self.token.privacyIDEAOtpLen)
        except ValueError:
            otplen = 6

        if getFromConfig("PrependPin") == "True" :
            pin = passw[0:-otplen]
            otpval = passw[-otplen:]
        else:
            pin = passw[otplen:]
            otpval = passw[0:otplen]

        #log.error("Pin: %s, otpval:%s",pin,otpval)

        return (res, pin, otpval)

    def checkPin(self, pin, options=None):
        '''
        checkPin - test is the pin is matching

        :param pin:      the pin
        :param options:  additional optional parameters, which could
                         be token specific
        :return: boolean

        '''
        res = False
        log.debug("entering checkPin function")

        if self.token.comparePin(pin) is True:
            res = True

        log.debug("result %r" % res)
        return res

    def statusValidationFail(self):
        #  callback to enable a status change, if auth failed
        return

    def statusValidationSuccess(self):
        #  callback to enable a status change, if auth failed
        return

    def __repr__(self):
        '''
        return the token state as text

        :return: token state as string representation
        :rtype:  string
        '''
        ldict = {}
        for attr in self.__dict__:
            key = "%r" % attr
            val = "%r" % getattr(self, attr)
            ldict[key] = val
        res = "<%r %r>" % (self.__class__, ldict)
        return res

    def get_vars(self, save=False):
        '''
        return the token state as dicts
        :return: token as dict
        '''
        ldict = {}
        for attr in self.__dict__:
            key = attr
            val = getattr(self, attr)
            if type(val) in [list, dict, str, unicode, int, float, bool]:
                ldict[key] = val
            elif type(val).__name__.startswith('Token'):
                ldict[key] = val.get_vars(save=save)
            else:
                ldict[key] = "%r" % val
        return ldict

    def getInitDetail(self, params, user=None):
        '''
        to complete the token normalisation, the response of the initialiastion
        should be build by the token specific method.
        
        getInitDetail returns additional information after an admin/init
        like the QR code of an HOTP/TOTP token.
        Can be anything else.
        '''
        response_detail = {}

        info = self.getInfo()
        response_detail.update(info)
        response_detail['serial'] = self.getSerial()

        otpkey = None
        if 'otpkey' in info:
            otpkey = info.get('otpkey')

        if otpkey is not None:
            response_detail["otpkey"] = {"description": _("OTP seed"),
                                         "value": "seed://%s" % otpkey,
                                         "img": create_img(otpkey, width=200)}

        return response_detail

    def getQRImageData(self, response_detail):
        '''
        '''
        url = None
        hparam = {}

        if response_detail is not None:
            if 'googleurl' in response_detail:
                url = response_detail.get('googleurl')
                hparam['alt'] = url

        return url, hparam


#### OcraTokenClass #####################################


class OcraTokenClass(TokenClass):
    '''
    OcraTokenClass  implement an ocra compliant token

    used from Config
        OcraMaxChallenges         - number of open challenges per token if None: 3
        OcraChallengeTimeout      - timeout definition like 1D, 2H or 3M if None: 1M
        OcraDefaultSuite          - if none :'OCRA-1:HOTP-SHA256-8:C-QN08'
        QrOcraDefaultSuite        - if none :'OCRA-1:HOTP-SHA256-8:C-QA64'


    algorithm Ocra Token Rollout: tow phases of rollout

    1. https://privacyideaserver/admin/init?
        type=ocra&
        genkey=1&
        sharedsecret=1&
        user=BENUTZERNAME&
        session=SESSIONKEY

        =>> "serial" : SERIENNUMMER, "sharedsecret" : DATAOBJECT, "app_import" : IMPORTURL
        - genSharedSecret - vom HSM oder urandom ?
        - app_import : + privacyidea://
                       + ocrasuite ->> default aus dem config: (DefaultOcraSuite)
                       + sharedsecret (Länge wie ???)
                       + seriennummer
        - seriennummer: uuid
        - token wird angelegt ist aber nicht aktiv!!! (counter == 0)


    2. https://privacyideaserver/admin/init?
        type=ocra&
        genkey=1&
        activationcode=AKTIVIERUNGSCODE&
        user=BENUTZERNAME&
        message=MESSAGE&
        session=SESSIONKEY

        =>> "serial" : SERIENNUMMER, "nonce" : DATAOBJECT, "transactionid" : "TRANSAKTIONSID, "app_import" : IMPORTURL

        - nonce - von HSM oder random ?
        - pkcs5 - kdf2
        - es darf zur einer Zeit nur eine QR Token inaktiv (== im Ausrollzustand) sein !!!!!
          der Token wird über den User gefunden
        - seed = pdkdf2(nonce + activcode + shared secret)
        - challenge generiern - von urandom oder HSM

    3. check_t
        - counter ist > nach der ersten Transaktion
        - if counter >= 1: delete sharedsecret löschen


    '''



    @classmethod
    def classInit(cls, param, user=None):

        helper_param = {}

        tok_type = "ocra"

        ## take the keysize from the ocrasuite
        ocrasuite = param.get("ocrasuite", None)
        activationcode = param.get("activationcode", None)
        sharedsecret = param.get("sharedsecret", None)
        serial = param.get("serial", None)
        genkey = param.get("genkey", None)

        if activationcode is not None:
            ## dont create a new key
            genkey = None
            serial = getRolloutToken4User(user=user, serial=serial, tok_type=tok_type)
            if serial is None:
                raise Exception('no token found for user: %r or serial: %r' % (user, serial))
            helper_param['serial'] = serial
            helper_param['activationcode'] = normalize_activation_code(activationcode)

        if ocrasuite is None:
            if sharedsecret is not None or  activationcode is not None:
                ocrasuite = getFromConfig("QrOcraDefaultSuite", 'OCRA-1:HOTP-SHA256-6:C-QA64')
            else:
                ocrasuite = getFromConfig("OcraDefaultSuite", 'OCRA-1:HOTP-SHA256-8:C-QN08')
            helper_param['ocrasuite'] = ocrasuite

        if genkey is not None:
            if ocrasuite.find('-SHA256'):
                key_size = 32
            elif ocrasuite.find('-SHA512'):
                key_size = 64
            else:
                key_size = 20
            helper_param['key_size'] = key_size

        return helper_param


    @classmethod
    @log_with(log)
    def getClassType(cls):
        '''
        getClassType - return the token type shortname

        :return: 'ocra'
        :rtype: string
        '''
        return "ocra"

    @classmethod
    @log_with(log)
    def getClassPrefix(cls):
        return "ocra"

    def __init__(self, aToken):
        '''
        getInfo - return the status of the token rollout

        :return: info of the ocra token state
        :rtype: dict
        '''
        TokenClass.__init__(self, aToken)
        self.setType(u"ocra")
        self.transId = 0
        return

    @log_with(log)
    def getInfo(self):
        '''
        getInfo - return the status of the token rollout

        :return: info of the ocra token state
        :rtype: dict
        '''
        return self.info

    @log_with(log)
    def update(self, params, reset_failcount=True):
        '''
        update: add further defintion for token from param in case of init
        '''
        if params.has_key('ocrasuite'):
            self.ocraSuite = params.get('ocrasuite')
        else:
            activationcode = params.get('activationcode', None)
            sharedSecret = params.get('sharedsecret', None)


            if activationcode is None and sharedSecret is None:
                self.ocraSuite = self.getOcraSuiteSuite()
            else:
                self.ocraSuite = self.getQROcraSuiteSuite()

        if params.get('activationcode', None):
            ## due to changes in the tokenclass parameter handling
            ## we have to add for compatibility a genkey parameter
            if params.has_key('otpkey') == False and params.has_key('genkey') == False:
                log.warning('missing parameter genkey\ to complete the rollout 2!')
                params['genkey'] = 1


        TokenClass.update(self, params, reset_failcount=reset_failcount)

        self.addToTokenInfo('ocrasuite', self.ocraSuite)

        ocraSuite = OcraSuite(self.ocraSuite)
        otplen = ocraSuite.truncation
        self.setOtpLen(otplen)

        ocraPin = params.get('ocrapin', None)
        if ocraPin is not None:
            self.token.setUserPin(ocraPin)

        if params.has_key('otpkey'):
            self.setOtpKey(params.get('otpkey'))

        self._rollout_1(params)
        self._rollout_2(params)

        return


    @log_with(log)
    def _rollout_1(self, params):
        '''
        do the rollout 1 step

        1. https://privacyideaserver/admin/init?
            type=ocra&
            genkey=1&
            sharedsecret=1&
            user=BENUTZERNAME&
            session=SESSIONKEY

            =>> "serial" : SERIENNUMMER, "sharedsecret" : DATAOBJECT, "app_import" : IMPORTURL
            - genSharedSecret - vom HSM oder urandom ?
            - app_import : + privacyidea://
                           + ocrasuite ->> default aus dem config: (DefaultOcraSuite)
                           + sharedsecret (Länge wie ???)
                           + seriennummer
            - seriennummer: uuid ??
            - token wird angelegt ist aber nicht aktiv!!! (counter == 0)

        '''
        sharedSecret = params.get('sharedsecret', None)
        if sharedSecret == '1':
            ##  preserve the rollout state
            self.addToTokenInfo('rollout', '1')

            ##  preseerver the current key as sharedSecret
            secObj = self.token.getHOtpKey()
            key = secObj.getKey()
            encSharedSecret = encryptPin(key)
            self.addToTokenInfo('sharedSecret', encSharedSecret)

            info = {}
            uInfo = {}

            info['sharedsecret'] = key
            uInfo['sh'] = key

            info['ocrasuite'] = self.getOcraSuiteSuite()
            uInfo['os'] = self.getOcraSuiteSuite()

            info['serial'] = self.getSerial()
            uInfo['se'] = self.getSerial()

            info['app_import'] = 'lseqr://init?%s' % (urllib.urlencode(uInfo))
            del info['ocrasuite']
            self.info = info

            self.token.privacyIDEAIsactive = False

        return

    @log_with(log)
    def _rollout_2(self, params):
        '''
        2.

        https://privacyideaserver/admin/init?
            type=ocra&
            genkey=1&
            activationcode=AKTIVIERUNGSCODE&
            user=BENUTZERNAME&
            message=MESSAGE&
            session=SESSIONKEY

        =>> "serial" : SERIENNUMMER, "nonce" : DATAOBJECT, "transactionid" : "TRANSAKTIONSID, "app_import" : IMPORTURL

        - nonce - von HSM oder random ?
        - pkcs5 - kdf2
        - es darf zur einer Zeit nur eine QR Token inaktiv (== im Ausrollzustand) sein !!!!!
          der Token wird über den User gefunden
        - seed = pdkdf2(nonce + activcode + shared secret)
        - challenge generiern - von urandom oder HSM

        '''
        activationcode = params.get('activationcode', None)
        if activationcode is not None:

            ##  genkey might have created a new key, so we have to rely on
            encSharedSecret = self.getFromTokenInfo('sharedSecret', None)
            if encSharedSecret is None:
                raise Exception ('missing shared secret of initialition for token %r' % (self.getSerial()))

            sharedSecret = decryptPin(encSharedSecret)

            ##  we generate a nonce, which in the end is a challenge
            nonce = createNonce()
            self.addToTokenInfo('nonce', nonce)

            ##  create a new key from the ocrasuite
            key_len = 20
            if self.ocraSuite.find('-SHA256'):
                key_len = 32
            elif self.ocraSuite.find('-SHA512'):
                key_len = 64


            newkey = kdf2(sharedSecret, nonce, activationcode, key_len)
            self.setOtpKey(binascii.hexlify(newkey))

            ##  generate challenge, which is part of the app_import
            message = params.get('message', None)
            (transid, challenge, _ret, url) = self.challenge(message)

            ##  generate response
            info = {}
            uInfo = {}
            info['serial'] = self.getSerial()
            uInfo['se'] = self.getSerial()
            info['nonce'] = nonce
            uInfo['no'] = nonce
            info['transactionid'] = transid
            uInfo['tr'] = transid
            info['challenge'] = challenge
            uInfo['ch'] = challenge
            if message is not None:
                uInfo['me'] = str(message.encode("utf-8"))


            ustr = urllib.urlencode({'u':str(url.encode("utf-8"))})
            uInfo['u'] = ustr[2:]
            info['url'] = str(url.encode("utf-8"))

            app_import = 'lseqr://nonce?%s' % (urllib.urlencode(uInfo))

            ##  add a signature of the url
            signature = {'si': self.signData(app_import) }
            info['signature'] = signature.get('si')

            info['app_import'] = "%s&%s" % (app_import, urllib.urlencode(signature))
            self.info = info

            ##  setup new state
            self.addToTokenInfo('rollout', '2')
            self.enable(True)

        return

    @log_with(log)
    def getOcraSuiteSuite(self):
        '''
        getQROcraSuiteSuite - return the QR Ocra Suite - if none, it will return the default

        :return: Ocrasuite of token
        :rtype: string
        '''
        defaultOcraSuite = getFromConfig("OcraDefaultSuite", 'OCRA-1:HOTP-SHA256-8:C-QN08')
        self.ocraSuite = self.getFromTokenInfo('ocrasuite', defaultOcraSuite)
        return self.ocraSuite

    @log_with(log)        
    def getQROcraSuiteSuite(self):
        '''
        getQROcraSuiteSuite - return the QR Ocra Suite - if none, it will return the default

        :return: QROcrasuite of token
        :rtype: string
        '''
        defaultOcraSuite = getFromConfig("QrOcraDefaultSuite", 'OCRA-1:HOTP-SHA256-8:C-QA64')
        self.ocraSuite = self.getFromTokenInfo('ocrasuite', defaultOcraSuite)
        return self.ocraSuite


    @log_with(log)
    def signData(self, data):
        '''
        sign the received data with the secret key

        :param data: arbitrary string object
        :type param: string

        :return: hexlified signature of the data
        '''
        secretHOtp = self.token.getHOtpKey()
        ocraSuite = OcraSuite(self.getOcraSuiteSuite(), secretHOtp)
        signature = ocraSuite.signData(data)
        return signature


    @log_with(log)
    def challenge(self, data, session='', typ='raw', challenge=None):
        '''
        the challenge method is for creating an transaction / challenge object

        remark: the transaction has a maximum lifetime and a reference to the OcraSuite token (serial)

        :param data:     data, which is the base for the challenge or None
        :type data:     string or None
        :param session:  session support for ocratokens
        :type session:  string
        :type typ:      define, which kind of challenge base should be used
                         could be raw - take the data input as is (extract chars accordind challenge definition Q)
                         or random    - will generate a random input
                         or hased     - will take the hash of the input data

        :return:    challenge response containing the transcation id and the challenge for the ocrasuite
        :rtype :    tuple of (transId(string), challenge(string))


        '''
        s_data = 'None'
        s_session = 'None'
        s_challenge = 'None'
        if data is not None: s_data = data
        if session is not None: s_session = session
        if challenge is None: s_challenge = challenge
        secretHOtp = self.token.getHOtpKey()
        ocraSuite = OcraSuite(self.getOcraSuiteSuite(), secretHOtp)

        if data is None or len(data) == 0:
            typ = 'random'

        if challenge is None:
            if typ == 'raw':
                challenge = ocraSuite.data2rawChallenge(data)
            elif typ == 'random':
                challenge = ocraSuite.data2randomChallenge(data)
            elif typ == 'hash':
                challenge = ocraSuite.data2hashChallenge(data)

        log.debug('challenge: %r ' % (challenge))

        serial = self.getSerial()
        counter = self.getOtpCount()

        ## set the pin onyl in the compliant hashed mode
        pin = ''
        if ocraSuite.P is not None:
            pinObj = self.token.getUserPin()
            pin = pinObj.getKey()

        try:
            param = {}
            param['C'] = counter
            param['Q'] = challenge
            param['P'] = pin
            param['S'] = session
            if ocraSuite.T is not None:
                now = datetime.datetime.now()
                stime = now.strftime("%s")
                itime = int(stime)
                param['T'] = itime

            ''' verify that the data is compliant with the OcraSuitesuite
                and the client is able to calc the otp
            '''
            c_data = ocraSuite.combineData(**param)
            ocraSuite.compute(c_data)

        except Exception as ex:
            log.error("%r" % (traceback.format_exc()))
            raise Exception('[OcraTokenClass] Failed to create ocrasuite challenge: %r' % (ex))

        ##  save the object
        digits = '0123456789'
        transid = ''
        transactionIdLen = 12

        try:
            transactionIdLen = int(getFromConfig("OcraDefaultSuite", '12'))
        except:
            transactionIdLen = 12
            log.debug("Failed to set transactionId length from config - using fallback %d" % (transactionIdLen))

        ##  create a non exisiting challenge
        try:
            while True:
                for _c in range(0, transactionIdLen):
                    transid += urandom.choice(digits)

                chall = OcraTokenClass.getTransaction(transid)
                if chall is None:
                    break

            ddata = ''
            if data is not None:
                ddata = data

            chall = OcraChallenge(transid, typ + ':' + challenge, serial, typ + ':' + ddata)
            chall.save()

        except Exception as ex:
            ##  this might happen if we have a db problem or the uniqnes constrain does not fit
            log.error("%r" % (traceback.format_exc()))
            raise Exception('[OcraTokenClass] Failed to create challenge object: %s' % (ex))

        realm = None
        realms = self.token.getRealms()
        if len(realms) > 0:
            realm = realms[0]

        url = ''
        if realm is not None:
            from privacyidea.lib.policy import PolicyClass
            from pylons import request, config, tmpl_context as c
            Policy = PolicyClass(request, config, c,
                                 get_privacyIDEA_config()) 
            url = Policy.get_qrtan_url(realm.name)

        return (transid, challenge, True, url)


    @log_with(log)
    def checkOtp(self, passw , counter , window , options=None):
        '''
        checkOtp - standard callback of privacyidea to verify the token

        :param passw:      the passw / otp, which has to be checked
        :type passw:       string
        :param counter:    the start counter
        :type counter:     int
        :param  window:    the window, in which the token is valid
        :type  window:     int
        :param options:    options contains the transaction id, eg. if check_t checks one transaction
                           this will support assynchreonous otp checks (when check_t is used)
        :type options:     dict

        :return:           verification counter or -1
        :rtype:            int (-1)

        '''
        ret = -1

        secretHOtp = self.token.getHOtpKey()
        ocraSuite = OcraSuite(self.getOcraSuiteSuite(), secretHOtp)

        ## if we have no transactionid given through the options,
        ## we have to retrieve the eldest challenge for this ocra token
        serial = self.getSerial()
        challenges = []

        ## set the ocra token pin
        ocraPin = ''
        if ocraSuite.P is not None:
            ocraPinObj = self.token.getUserPin()
            ocraPin = ocraPinObj.getKey()

            if ocraPin is None or len(ocraPin) == 0:
                ocraPin = ''

        timeShift = 0
        if  ocraSuite.T is not None:
            defTimeWindow = int(getFromConfig("ocra.timeWindow", 180))
            window = int(self.getFromTokenInfo('timeWindow', defTimeWindow)) / ocraSuite.T
            defTimeShift = int(getFromConfig("ocra.timeShift", 0))
            timeShift = int(self.getFromTokenInfo("timeShift", defTimeShift))

        if options is None:
            challenges = OcraTokenClass.getTransactions4serial(serial, currentOnly=True)

        elif options is not None:
            if type(options).__name__ != 'dict':
                err = '[chekOtp] "options" not of type dict! %r' % (type(options))
                log.error(err)
                raise Exception(err)

            if options.has_key('transactionid'):
                transid = options.get('transactionid')
                challenges.append(OcraTokenClass.getTransaction(transid))

            elif options.has_key('challenge'):
                challenges.append(options)

            ## due to the added options in checkUserPass, we have to extend
            ## the logic here:
            ## if no challenges found in between but we have a serial, we catch
            ## the open challenges by serial (s.o.)
            if len(challenges) == 0:
                challenges = OcraTokenClass.getTransactions4serial(serial, currentOnly=True)

        if len(challenges) == 0:
            ##  verify that there has already been a challenge
            challenges = OcraTokenClass.getTransactions4serial(serial)
            if len(challenges) > 0:
                err = 'No current transaction found!'
                ret = -1
                return ret
            else:
                err = 'No open transaction found!'
                log.error(err)
                if type(options) == dict and options.has_key('transactionid'):
                    raise Exception(err)
                ret = -1
                return ret

        for ch in challenges:
            challenge = {}

            if type(ch) == dict:
                ##  transaction less checkOtp
                self.transId = 0
                challenge.update(ch)

            elif type(ch) == OcraChallenge:
                ##  preserve transaction context, so we could use this in the status callback
                self.transId = ch.transid
                challenge['challenge'] = ch.challenge
                challenge['transid'] = ch.transid
                challenge['session'] = ch.session



            ret = ocraSuite.checkOtp(passw, counter, window, challenge, pin=ocraPin , options=options, timeshift=timeShift)

            if ret != -1:
                break

        if -1 == ret:
            ##  autosync: test if two consecutive challenges + it's counter match
            ret = self.autosync(ocraSuite, passw, challenge)

        return ret

    @log_with(log)
    def autosync(self, ocraSuite, passw, challenge):
        '''
        try to resync a token automaticaly, if a former and the current request failed

        :param  ocraSuite: the ocraSuite of the current Token
        :type  ocraSuite: ocra object
        :param  passw:
        '''
        res = -1

        autosync = False

        try:
            async = getFromConfig("AutoResync")
            if async is None:
                autosync = False
            elif "true" == async.lower():
                autosync = True
            elif "false" == async.lower():
                autosync = False
        except Exception as ex:
            log.error('autosync check undefined %r' % (ex))
            return res

        ' if autosync is not enabled: do nothing '
        if False == autosync:
            return res

        ##
        ## AUTOSYNC starts here
        ##

        counter = self.token.getOtpCounter()
        syncWindow = self.token.getSyncWindow()
        if  ocraSuite.T is not None:
            syncWindow = syncWindow / 10


        ## set the ocra token pin
        ocraPin = ''
        if ocraSuite.P is not None:
            ocraPinObj = self.token.getUserPin()
            ocraPin = ocraPinObj.getKey()

            if ocraPin is None or len(ocraPin) == 0:
                ocraPin = ''

        timeShift = 0
        if  ocraSuite.T is not None:
            timeShift = int(self.getFromTokenInfo("timeShift", 0))

        #timeStepping    = int(ocraSuite.T)

        tinfo = self.getTokenInfo()

        ## autosync does only work, if we have a token info, where the last challenge and the last sync-counter is stored
        ## if no tokeninfo, we start with a autosync request, thus start the lookup in the sync window

        if tinfo.has_key('lChallenge') == False:
            ## run checkOtp, with sync window for the current challenge
            log.debug('initial sync')
            count_0 = -1
            try:
                otp0 = passw
                count_0 = ocraSuite.checkOtp(otp0, counter, syncWindow, challenge, pin=ocraPin, timeshift=timeShift)
            except Exception as ex:
                log.error(' error during autosync0 %r' % (ex))

            if count_0 != -1:
                tinfo['lChallenge'] = {'otpc' : count_0}
                self.setTokenInfo(tinfo)
                log.info('initial sync - success: %r' % (count_0))

            res = -1
            log.debug('initial sync done!')

        else:
            ## run checkOtp, with sync window for the current challenge
            log.debug('sync')
            count_1 = -1
            try:
                otp1 = passw
                count_1 = ocraSuite.checkOtp(otp1, counter, syncWindow, challenge, pin=ocraPin, timeshift=timeShift)
            except Exception as ex:
                log.error(' error during autosync1 %r' % (ex))

            if count_1 == -1:
                del tinfo['lChallenge']
                self.setTokenInfo(tinfo)
                log.warn('sync failed! Not a valid pass in scope (%r)' % (otp1))
                res = -1
            else:
                ## run checkOtp, with sync window for the old challenge
                lChallange = tinfo.get('lChallenge')
                count_0 = lChallange.get('otpc')

                if ocraSuite.C is not None:
                    ##  sync the counter based ocra token
                    if count_1 - count_0 < 2:
                        self.setOtpCount(count_1)
                        res = count_1

                if ocraSuite.T is not None:
                    ##  sync the timebased ocra token
                    if count_1 - count_0 < ocraSuite.T * 2 :
                        ## calc the new timeshift !
                        log.debug("the counter %r matches: %r" %
                                  (count_1, datetime.datetime.fromtimestamp(count_1)))

                        currenttime = int(time.time())
                        new_shift = (count_1 - currenttime)

                        tinfo['timeShift'] = new_shift
                        self.setOtpCount(count_1)
                        res = count_1

                ##  if we came here, the old challenge is not required anymore
                del tinfo['lChallenge']
                self.setTokenInfo(tinfo)

            log.debug('sync done!')

        return res

    def is_challenge_response(self, passw, user, options=None, challenges=None):
        '''
        check, if the request contains the result of a challenge

        :param passw: password, which might be pin or pin+otp
        :param user: the requesting user
        :param options: dictionary of additional request parameters

        :return: returns true or false
        '''

        challenge_response = False

        return challenge_response

    @log_with(log)
    def statusValidationFail(self):
        '''
        statusValidationFail - callback to enable a status change,

        will be called if the token verification has failed

        :return - nothing

        '''
        ocraChallenge = None

        if self.transId == 0 :
            return

        try:
            ocraChallenge = OcraTokenClass.getTransaction(self.transId)
            ocraChallenge.setTanStatus(received=True, valid=False)

            ##  still in rollout state??
            rolloutState = self.getFromTokenInfo('rollout', '0')

            if rolloutState == '1':
                log.info('rollout state 1 for token %r not completed' % (self.getSerial()))

            elif rolloutState == '2':
                try:
                    maxchall = int(getFromConfig("OcraMaxChallengeRequests", '3'))
                except:
                    maxchall = 3

                if int(ocraChallenge.received_count) >= maxchall:
                    ##  after 3 fails in rollout state 2 - reset to rescan
                    self.addToTokenInfo('rollout', '1')
                    log.info('rollout for token %r reset to phase 1:' % (self.getSerial()))

                log.info('rollout for token %r not completed' % (self.getSerial()))

        except Exception as ex:
            log.error('Error during validation finalisation for token %r :%r' % (self.getSerial(), ex))
            log.error("%r" % (traceback.format_exc()))
            raise Exception(ex)

        finally:
            if ocraChallenge != None:
                ocraChallenge.save()

        return

    
    @log_with(log)
    def statusValidationSuccess(self):
        '''
        statusValidationSuccess - callback to enable a status change,

        remark: will be called if the token shas been succesfull verified

        :return: - nothing

        '''
        if self.transId == 0 :
            return

        ocraChallenge = OcraTokenClass.getTransaction(self.transId)
        ocraChallenge.setTanStatus(received=True, valid=True)
        ocraChallenge.save()

        ##  still in rollout state??
        rolloutState = self.getFromTokenInfo('rollout', '0')

        if rolloutState == '2':
            t_info = self.getTokenInfo()
            if t_info.has_key('rollout'):
                del t_info['rollout']
            if t_info.has_key('sharedSecret'):
                del t_info['sharedSecret']
            if t_info.has_key('nonce'):
                del t_info['nonce']
            self.setTokenInfo(t_info)

            log.info('rollout for token %r completed' % (self.getSerial()))

        elif rolloutState == '1':
            raise Exception('unable to complete the rollout ')
        return


    @log_with(log)
    def resync(self, otp1, otp2, options=None):
        '''
        - for the resync to work, we take the last two transactions and their challenges
        - for each challenge, we search forward the sync window length

        '''
        ret = False
        challenges = []

        o_challenges = OcraTokenClass.getTransactions4serial(self.getSerial())
        for challenge in o_challenges:
            challenges.append(challenge)

        ##  check if there are enough challenges around
        if len(challenges) < 2:
            return False

        challenge1 = {}
        challenge2 = {}

        if options is None:
            ch1 = challenges[0]
            challenge1['challenge'] = ch1.challenge
            challenge1['transid'] = ch1.transid
            challenge1['session'] = ch1.session

            ch2 = challenges[1]
            challenge2['challenge'] = ch2.challenge
            challenge2['transid'] = ch2.transid
            challenge2['session'] = ch2.session

        else:
            if options.has_key('challenge1'):
                challenge1['challenge'] = options.get('challenge1')
            if options.has_key('challenge2'):
                challenge2['challenge'] = options.get('challenge2')


        if len(challenge1) == 0 or len(challenge2) == 0:
            error = "No challeges found!"
            log.error('%s' % (error))
            raise Exception('[OcraTokenClass:resync] %s' % (error))



        secretHOtp = self.token.getHOtpKey()
        ocraSuite = OcraSuite(self.getOcraSuiteSuite(), secretHOtp)

        syncWindow = self.token.getSyncWindow()
        if  ocraSuite.T is not None:
            syncWindow = syncWindow / 10

        counter = self.token.getOtpCounter()

        ## set the ocra token pin
        ocraPin = ''
        if ocraSuite.P is not None:
            ocraPinObj = self.token.getUserPin()
            ocraPin = ocraPinObj.getKey()

            if ocraPin is None or len(ocraPin) == 0:
                ocraPin = ''

        timeShift = 0
        if  ocraSuite.T is not None:
            timeShift = int(self.getFromTokenInfo("timeShift", 0))

        try:

            count_1 = ocraSuite.checkOtp(otp1, counter, syncWindow, challenge1, pin=ocraPin, timeshift=timeShift)
            if count_1 == -1:
                log.info('lookup for first otp value failed!')
                ret = False
            else:
                count_2 = ocraSuite.checkOtp(otp2, counter, syncWindow, challenge2, pin=ocraPin, timeshift=timeShift)
                if count_2 == -1:
                    log.info('lookup for second otp value failed!')
                    ret = False
                else:
                    if ocraSuite.C is not None:
                        if count_1 + 1 == count_2:
                            self.setOtpCount(count_2)
                            ret = True

                    if  ocraSuite.T is not None:
                        if count_1 - count_2 <= ocraSuite.T * 2:
                            ##  callculate the timeshift
                            date = datetime.datetime.fromtimestamp(count_2)
                            log.info('syncing token to new timestamp: %r' % (date))

                            now = datetime.datetime.now()
                            stime = now.strftime("%s")
                            timeShift = count_2 - int(stime)
                            self.addToTokenInfo('timeShift', timeShift)
                            ret = True

        except Exception as ex:
            log.error('unknown error: %r' % (ex))
            raise Exception('[OcraTokenClass:resync] unknown error: %s' % (ex))

        return ret


    @log_with(log)
    def getStatus(self, transactionId):
        '''
        getStatus - assembles the status of a transaction / challenge in a dict

        {   "serial": SERIENNUMMER1,
            "transactionid": TRANSACTIONID1,
            "received_tan": true,
            "valid_tan": true,
            "failcount": 0
        }

        :param transactionId:    the transaction / challenge id
        :type transactionId:    string

        :return:    status dict
        :rtype:       dict
        '''
        statusDict = {}
        ocraChallenge = OcraTokenClass.getTransaction(transactionId)
        if ocraChallenge is not None:
            statusDict['serial'] = ocraChallenge.tokenserial
            statusDict['transactionid'] = ocraChallenge.transid
            statusDict['received_tan'] = ocraChallenge.received_tan
            statusDict['valid_tan'] = ocraChallenge.valid_tan
            statusDict['failcount'] = self.getFailCount()
            statusDict['id'] = ocraChallenge.id
            statusDict['timestamp'] = unicode(ocraChallenge.timestamp)
            statusDict['active'] = unicode(self.isActive())

        return statusDict


    @classmethod
    @log_with(log)
    def timeoutJanitor(cls):
        '''
        timeoutJanitor - remove all outdated transactions / challenges

        :return: - nothing

        '''
        delta = datetime.timedelta(days=0)
        scopeDef = getFromConfig("OcraChallengeTimeout", '1D')

        ##  timedelta supports : days[, seconds[, microseconds[, milliseconds[, minutes[, hours[, weeks]]]]]]])
        if re.match('^(\d+[DHMS])+$', scopeDef):
            delta = datetime.timedelta(days=0)
            parts = re.findall('\d+[DHMS]', scopeDef)
            for part in parts:
                period = part[-1]
                quantity = int(part[:-1])
                if period == 'D':
                    delta = delta + datetime.timedelta(days=quantity)
                elif period == 'H':
                    delta = delta + datetime.timedelta(hours=quantity)
                elif period == 'M':
                    delta = delta + datetime.timedelta(minutes=quantity)
                elif period == 'S':
                    delta = delta + datetime.timedelta(seconds=quantity)
        else:
            log.info('OcraChallengeTimeout value %r does not match timedelta definition (^(\d+[DHMS])+$)' % (scopeDef))
            try:
                scope_def = int(scopeDef)
                delta = datetime.timedelta(seconds=scope_def)
            except ValueError:
                log.info('Failed to convert OcraChallengeTimeout value from config: %r' % (scopeDef))
                delta = datetime.timedelta(days=1)


        ocraChallenges = Session.query(OcraChallenge).filter(
                            OcraChallenge.timestamp < datetime.datetime.now() - delta)


        for ocraChallenge in ocraChallenges:
            log.warning("dropping outdated ocraChallenge): %r for token %r" %
                        (ocraChallenge.transid, ocraChallenge.tokenserial))
            Session.delete(ocraChallenge)

        return

    @classmethod
    @log_with(log)
    def maxChallengeRequestJanitor(cls):
        '''
        maxChallengeRequestJanitor - remove all transactions / challenges which have been made more than maxChallengeRequests

        :return: - nothing

        '''
        maxRequests = int(getFromConfig("OcraMaxChallengeRequests", '3'))

        ocraChallenges = Session.query(OcraChallenge).filter(
                            OcraChallenge.received_count >= maxRequests)


        for ocraChallenge in ocraChallenges:
            log.warning("dropping outdated ocraChallenge): %r for token %r" % (ocraChallenge.transid, ocraChallenge.tokenserial))
            Session.delete(ocraChallenge)
        return

    @classmethod
    @log_with(log)
    def maxChallengeJanitor(cls, transId=None, serial=None):
        '''
        maxChallengeJanitor - remove for one token (serial) all challengens but the last ones

        :param transId:     the current transaction, which provides a the lookup for the serial number
        :type transId:     string

        :param serial:     the serial number of the token
        :type serial:     string

        :return: - nothing

        '''
        maxChallDef = getFromConfig("OcraMaxChallenges", '3')
        try:
            ones = int(maxChallDef)
        except ValueError as ex:
            log.error('Faild to convert OcraMaxChallenges value from config: %r :%r' % (maxChallDef, ex))
            ones = 3

        if ones <= 0:
            ones = 3

        if transId is not None:
            challenges = Session.query(OcraChallenge).filter(OcraChallenge.transid == u'' + transId)
            if challenges is None:
                log.info('no ocraChallenge found for tranid %r' % (transId))
                return

            for challenge in challenges:
                serial = challenge.tokenserial

        if serial is None:
            log.error('failed to lookup for transid %r or serial %r' % (transId, serial))
            return

        challenges = Session.query(OcraChallenge).\
                        filter(OcraChallenge.tokenserial == u'' + serial)\
                        .order_by(desc(OcraChallenge.id))

        lastIds = set()
        for challenge in challenges:
            if len(lastIds) < ones:
                lastIds.add(challenge.id)
            else:
                log.warning("dropping max ocraChallenges: %r :: %r for token %r" % \
                            (challenge.id, challenge.transid, challenge.tokenserial))
                Session.delete(challenge)
        return

    @classmethod
    @log_with(log)
    def getTransaction(cls, transId):
        '''
        getTransaction - lookup for the challenge object of the given id

        :param transId:   challenge identifier
        :type transId:   string

        :return: the challenge data object
        :rtype: OcraChallenge

        '''
        ##  first do housekeeping - remove outdated transactions
        cls.timeoutJanitor()
        cls.maxChallengeRequestJanitor()
        cls.maxChallengeJanitor(transId=transId)

        ocraChallenge = None
        count = 0


        if transId is not None:
            challenges = Session.query(OcraChallenge).filter(OcraChallenge.transid == u'' + transId)


        if challenges is None:
            log.info('no ocraChallenge found for tranid %r' % (transId))
            return None


        for ocraChallenge in challenges:
            log.debug("%r for token: %r" % (ocraChallenge.transid, ocraChallenge.tokenserial))
            count += 1

        if count == 0 or count > 1:
            log.error('%r ocraChallenge token found for this transaction %r ' % (count, transId))
            #raise Exception('%r ocraChallenge token found for this transaction %r '%(count,transId))

        log.debug('%r' % (ocraChallenge))
        return ocraChallenge


    @classmethod
    @log_with(log)
    def getTransactions4serial(cls, serial, currentOnly=False):
        '''
        getTransactions4serial - give all challenges for a given token serial number

        :param serial:     token serial identifier
        :type serial:     string
        :param currentOnly: boolean Flag to return all Challenges (like for status request)
                             or to return the eldest open transaction / challenge
        :type currentOnly: boolean flag

        :return:         return a list of Challenges
        :rtype:         OcraChallenge obejct list

        '''
        log.debug('%r: %r' % (serial, currentOnly))

        ##  first do housekeeping - remove outdated transactions
        cls.timeoutJanitor()
        cls.maxChallengeRequestJanitor()
        cls.maxChallengeJanitor(serial=serial)


        ocraChallenges = []
        ocraChallenge = None
        challenges = []


        if serial is not None:
            if currentOnly == False:
                challenges = Session.query(OcraChallenge)\
                    .filter(OcraChallenge.tokenserial == u'' + serial)\
                    .order_by(desc(OcraChallenge.id))
            else:
                ##  return the oldest transaction onyl -  orderby(id).limit(1)
                challenges = Session.query(OcraChallenge)\
                    .filter(OcraChallenge.tokenserial == u'' + serial)\
                    .filter(OcraChallenge.received_tan == False)\
                    .order_by(asc(OcraChallenge.id))

        if challenges is None:
            log.info('no ocraChallenge found for serial %r' % (serial))
            return None

        for ocraChallenge in challenges:
            log.debug("%r for token: %r" % (ocraChallenge.transid, ocraChallenge.tokenserial))
            ocraChallenges.append(ocraChallenge)

        return ocraChallenges

    def getInitDetail(self, params , user=None):
        '''
        to complete the token normalisation, the response of the initialiastion
        should be build by the token specific method, the getInitDetails
        '''
        response_detail = {}

        info = self.getInfo()
        response_detail.update(info)

        otpkey = None
        if 'otpkey' in info:
            otpkey = info.get('otpkey')
        response_detail["otpkey"] = otpkey

        ocra_url = info.get('app_import')
        response_detail["ocraurl"] = {
               "description": _("URL for OCRA token"),
               "value": ocra_url,
                "img": create_img(ocra_url, width=250),
               }

        return response_detail

    def getQRImageData(self, response_detail):
        '''
        '''
        url = None
        hparam = {}

        if response_detail is not None:
            if 'ocraurl' in response_detail:
                url = response_detail.get('ocraurl', {}).get('value', '')
                hparam['alt'] = response_detail.get('app_import', '')
        return url, hparam


##eof##########################################################################



# -*- coding: utf-8 -*-
#
#  privacyIDEA is a fork of LinOTP
#  May 08, 2014 Cornelius Kölbel
#  License:  AGPLv3
#  contact:  http://www.privacyidea.org
#
#  Copyright (C) 2010 - 2014 LSE Leading Security Experts GmbH
#  License:  AGPLv3
#  contact:  http://www.linotp.org
#            http://www.lsexperts.de
#            linotp@lsexperts.de
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
'''
  Description:  This module can be used to create the optical challenge image
                for Feitian c601 token.
                
                c601:
                Either you can create a single challenge that will display
                the OTP response
                    calculate_optical_data(value="12345678")
                
                or you can split the challenge into ACCOUNT and AMOUNT.
                     calculate_optical_data(value="12345", type=ACCOUNT)
                     calculate_optical_data(value="67890", type=AMOUNT)
                     
                The challenge then will be AMOUNT+ACCOUNT (6789012345)
'''
import os

CHALLENGE = [0, 1]
ACCOUNT = [1, 1]
AMOUNT = [1, 0]

DELAY_BLACK = 5
DELAY_WHITE = 5
DELAY_GRAY = 5


def _calculate_sum(value):
    summ = 0
    alternate = True

    for c in value:
        n = ord(c) & 0x0f

        if alternate:
            v = n * 2
            if v > 9:
                summ += 1 + (v % 10)
            else:
                summ += v
            alternate = False
        else:
            summ += n
            alternate = True

    return (10 - (summ % 10)) % 10


def _modify_optical(optical):
    '''
    Takes the optical array and modifies it to the three
        black = 0
        gray =  1
        white = 2
        values
    '''
    ret_optical = []
    t = 0
    ret_optical.append(t)

    for o in optical[:-1]:
        if o == 1:
            t = t + 1
            if t > 2:
                t = 0
        else:
            t = t - 1
            if t < 0:
                t = 2

        ret_optical.append(t)

    return ret_optical


def calculate_optical_data(value="", type=CHALLENGE):
    ''' create an array, representing black and white (1 and 0)
    
    type
        0,1 : Default challenge
        1,1 : Account
        1,0 : Amount
        
    value
        Input challenge Data: A numerical string
    '''

    # sanity checks
    if len(type) != 2:
        raise Exception("The type is limited to a length of 2.")

    _value_int = int(value)

    if type == CHALLENGE:
        if len(value) > 59:
            raise Exception("The length of the challenge value is"
                            " limited to 59.")

        if len(value) < 8:
            raise Exception("The length of the challenge value must"
                            " be a minimum of 8.")

    if type == ACCOUNT or type == AMOUNT:
        if len(value) > 6:
            raise Exception("You should not use amount or account longer "
                            "than 6, as it will not be displayerd.")

    optical = []

    # header
    optical.append(1)
    optical.append(1)
    optical.append(1)
    optical.append(1)
    optical.append(1)
    optical.append(0)

    # type
    optical.append(type[0])
    optical.append(type[1])

    # Length
    L = "%02d" % len(value)

    # convert a "9" to a binary 1001
    # Lenght
    for c in L:
        optical.append(ord(c) >> 3 & 0x01)
        optical.append(ord(c) >> 2 & 0x01)
        optical.append(ord(c) >> 1 & 0x01)
        optical.append(ord(c) & 0x01)

    # data
    for v in value:
        optical.append(ord(v) >> 3 & 0x01)
        optical.append(ord(v) >> 2 & 0x01)
        optical.append(ord(v) >> 1 & 0x01)
        optical.append(ord(v) & 0x01)

    # checksum
    # sum 0011 - 3 0010 - 2 0001 - 1
    bin_data = int("%d%d" % (type[0], type[1]))
    bin_str = "%d" % bin_data

    # type + len + value
    sum_data = bin_str + L + value

    summ = _calculate_sum(sum_data)

    optical.append(summ >> 3 & 0x01)
    optical.append(summ >> 2 & 0x01)
    optical.append(summ >> 1 & 0x01)
    optical.append(summ & 0x01)

    optical = _modify_optical(optical)

    return optical


def create_image(optical,
                 outfile="challenge.gif",
                 delay=(DELAY_WHITE, DELAY_GRAY, DELAY_BLACK)):
    '''
    Creates an animated GIF for the given data array
    which is the output of calculate optical data
    '''
    input_files = ""
    for o in optical:
        if o == 2:
            input_files += " -delay %d white.gif " % delay[0]
        elif o == 1:
            input_files += " -delay %d gray.gif " % delay[1]
        elif o == 0:
            input_files += " -delay %d black.gif  " % delay[2]
    command = "convert  %s  -loop 0 %s" % (input_files, outfile)
    os.system(command)

if __name__ == '__main__':

    # data for challenge 12345678
    demo_data_challenge = [0, 1, 2, 0, 1, 2, 1, 0, 1, 0, 2, 1, 0, 1, 0, 2,
                           1, 0, 2, 1, 2, 1, 0, 1, 0, 2, 1, 2, 0, 2, 0, 2,
                           1, 0, 1, 0, 1, 0, 1, 2, 1, 0, 1, 2, 0, 1, 0, 2,
                           1, 0, 2, 0]
    optical = calculate_optical_data(value="12345678")
    print "-" * 70
    print "Cha D: %s" % demo_data_challenge
    print "Cha O: %s " % optical
    create_image(optical)

    # data for account 12345
    demo_data_account = [0, 1, 2, 0, 1, 2, 1, 2, 0, 2, 1, 0, 2, 1, 2, 1, 2,
                         1, 0, 2, 0, 2, 1, 2, 1, 0, 2, 0, 1, 0, 1, 0, 2, 1,
                         2, 1, 2, 1, 2, 0]
    optical = calculate_optical_data(value="12345", type=ACCOUNT)
    print "-" * 70
    print "Acc D: %s" % demo_data_account
    print "Acc O: %s " % optical
    create_image(optical, outfile="account.gif")

    # data for amount
    demo_data_amount = [0, 1, 2, 0, 1, 2, 1, 2, 1, 0, 2, 1, 0, 2, 0, 2, 0, 2,
                        0, 1, 0, 2, 0, 1, 2, 0, 2, 1, 0, 1, 0, 2, 0, 2, 1, 0,
                        2, 1, 2, 0]
    optical = calculate_optical_data(value="67890", type=AMOUNT)
    print "-" * 70
    print "Amo D: %s" % demo_data_amount
    print "Amo O: %s " % optical
    create_image(optical, outfile="amount.gif")



# -*- coding: utf-8 -*-
#
#  privacyIDEA is a fork of LinOTP
#  May 08, 2014 Cornelius Kölbel
#  License:  AGPLv3
#  contact:  http://www.privacyidea.org
#
#  Copyright (C) 2010 - 2014 LSE Leading Security Experts GmbH
#  License:  LSE
#  contact:  http://www.linotp.org
#            http://www.lsexperts.de
#            linotp@lsexperts.de
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
'''
  Description:  the security provider is a dynamic hanlder for the security
                relevant tasks like

                random 
                crypt
                decrypt
                sign

  Dependencies: -

'''


import thread
import time
import logging
import traceback

from privacyidea.lib.crypto import zerome
from privacyidea.lib.error import HSMException

DEFAULT_KEY = 0
CONFIG_KEY = 1
TOKEN_KEY = 2
VALUE_KEY = 3


log = logging.getLogger(__name__)

class SecurityProvider(object):
    '''
    the Security provider is the singleton in the server who provides
    the security modules to run security relevant methods

    - read the hsm configurations
    - set up a pool of hsm modules
    - bind a hsm to one session
    - free the hsm from session after usage

    as session identifier the thread id is used
    '''
    def __init__(self, secLock):
        '''
        setup the SecurityProvider, which is called on server startup
        from the app_globals init

        :param secLock: RWLock() to support server wide locking
        :type  secLock: RWLock

        :return: -

        '''
        self.config = {}
        self.security_modules = {}
        self.activeOne = 'default'
        self.hsmpool = {}
        self.rwLock = secLock
        self.max_retry = 5

    def __createDefault__(self, config):
        ''' 
        create a backward compatible default security provider

        :param config:

        '''
        keyFile = None
        if config.has_key('privacyideaSecretFile'):
            keyFile = config.get('privacyideaSecretFile')
        self.config['default'] = {'pinHandle'   : TOKEN_KEY,
                                  'passHandle'  : CONFIG_KEY,
                                  'valueHandle' : VALUE_KEY,
                                  'defaultHandle' : DEFAULT_KEY,
                                  'crypted'     : 'FALSE',
                                  'file'        : keyFile,
                                  'module'      : 'privacyidea.lib.security.default.DefaultSecurityModule',
                                  'poolsize'    : 20,
                                  }

        self.config['err'] = {'pinHandle'   : TOKEN_KEY,
                                  'passHandle'  : CONFIG_KEY,
                                  'valueHandle' : VALUE_KEY,
                                  'defaultHandle' : DEFAULT_KEY,
                                  'crypted'     : 'FALSE',
                                  'file'        : keyFile,
                                  'module'      : 'privacyidea.lib.security.default.ErrSecurityModule',
                                  'poolsize'    : 20,
                                  }

    def load_config(self, config):
        '''
        load the security modules configuration
        '''

        try:
            ## load backward compatible defaults
            self. __createDefault__(config)

            for key in config:

                ## lookup, which is the active security module
                if key == 'privacyideaActiveSecurityModule':
                    self.activeOne = config.get(key)
                    log.debug("setting active security module: %s" % self.activeOne)

                if key.startswith('privacyideaSecurity'):
                    entry = key.replace('privacyideaSecurity.', '')
                    try:
                        (id, val) = entry.split('.')
                    except Exception as e:
                        error = ('[SecurityProvider:load_config] failed to '
                                 'identify config entry: %s ' % (unicode(key)))
                        log.error(error)
                        log.error("%s" % traceback.format_exc())
                        raise HSMException(error, id=707)

                    if self.config.has_key(id):
                        id_config = self.config.get(id)
                        id_config[val] = config.get(key)
                    else:
                        self.config[id] = {val:config.get(key) }

        except Exception as e:
            log.error("failed to identify module: %r " % e)
            error = "failed to identify module: %s " % unicode(e)
            raise HSMException(error, id=707)

        ## now create for each module a pool of hsm objects
        self.rwLock.acquire_write()
        try:
            for id in self.config:
                self.createHSMPool(id)
        finally:
            self.rwLock.release()

        return


    def loadSecurityModule(self, id=None):
        ''' 
        return the specified security module 
        
        :param id:  identifier for the security module (from the configuration)
        :type  id:  String or None
        
        :return:    None or the created object
        :rtype:     security module
        '''

        ret = None

        if id is None:
            id = self.activeOne

        log.debug("Loading module %s" % id)

        if self.config.has_key(id) == False:
            return ret

        config = self.config.get(id)
        if config.has_key('module') == False:
            return ret

        module = config.get('module')
        methods = ["encrypt", "decrypt", "random", "setup_module"]
        method = ""

        parts = module.split('.')
        className = parts[-1]
        packageName = '.'.join(parts[:-1])

        mod = __import__(packageName, globals(), locals(), [className])
        klass = getattr(mod, className)

        for method in methods:
            if hasattr(klass, method) == False:
                error = ("[loadSecurityModule] Security Module %s misses the "
                         "following interface: %s" % (unicode(module), unicode(method)))
                log.error(error)
                raise NameError(error)

        ret = klass(config)
        self.security_modules[id] = ret

        log.debug("returning %r" % ret)

        return ret


    def _getHsmPool_(self, hsm_id):
        ret = None
        if self.hsmpool.has_key(hsm_id):
            ret = self.hsmpool.get(hsm_id)
        return ret


    def setupModule(self, hsm_id, config=None):
        ''' 
        setupModule is called during runtime to define
        the config parameters like passw or connection strings
        '''
        self.rwLock.acquire_write()
        try:
            pool = self._getHsmPool_(hsm_id)
            if pool is None:
                error = ("[setupModule] failed to retieve pool "
                         "for hsm_id: %s" % (unicode(hsm_id)))
                log.error(error)
                raise HSMException(error, id=707)

            for entry in pool:
                hsm = entry.get('obj')
                hsm.setup_module(config)

            self.activeOne = hsm_id
        except Exception as e:
            error = "[setupModule] failed to load hsm : %s" % (unicode(e))
            log.error(error)
            log.error(traceback.format_exc())
            raise HSMException(error, id=707)

        finally:
            self.rwLock.release()
        return self.activeOne

    def createHSMPool(self, hsm_id=None, *args, **kw):
        '''
        setup a pool of secutity provider
        '''
        pool = None
        ## amount has to be taken from the hsm-id config
        if hsm_id == None:
            ids = self.config
        else:
            if self.config.has_key(hsm_id):
                ids = []
                ids.append(hsm_id)
            else:
                error = "[createHSMPool] failed to find hsm_id: %s" % (unicode(hsm_id))
                log.error(error)
                raise HSMException(error, id=707)

        for id in ids:
            pool = self._getHsmPool_(id)
            log.debug("already got this pool: %r" % pool)
            if pool is None:
                ## get the number of entries from the hsd (id) config
                conf = self.config.get(id)
                amount = int(conf.get('poolsize', 10))
                log.debug("creating pool for %r with size %r" % (id, amount))
                pool = []
                for i in range(0, amount):
                    error = ''
                    hsm = None
                    try:
                        hsm = self.loadSecurityModule(id)
                    except Exception as e:
                        error = u"%r" % e
                        log.error("%r" % (e))
                        log.error(traceback.format_exc())
                    pool.append({'obj': hsm , 'session': 0, 'error':error})

                self.hsmpool[id] = pool
        return pool

    def _findHSM4Session(self, pool, sessionId):
        found = None
        ## find session
        for hsm in pool:
            hsession = hsm.get('session')
            if hsession == sessionId:
                found = hsm
        return found

    def _createHSM4Session(self, pool, sessionId):
        found = None
        for hsm in pool:
            hsession = hsm.get('session')
            if unicode(hsession) == u'0':
                hsm['session'] = sessionId
                found = hsm
                break
        return found

    def _freeHSMSession(self, pool, sessionId):
        hsm = None
        for hsm in pool:
            hsession = hsm.get('session')
            if unicode(hsession) == unicode(sessionId):
                hsm['session'] = 0
                break
        return hsm

    def dropSecurityModule(self, hsm_id=None, sessionId=None):
        found = None
        if hsm_id is None:
            hsm_id = self.activeOne
        if sessionId is None:
            sessionId = unicode(thread.get_ident())

        if self.config.has_key(hsm_id) == False:
            error = ('[SecurityProvider:dropSecurityModule] no config found '
                     'for hsm with id %s ' % (unicode(hsm_id)))
            log.error(error)
            raise HSMException(error, id=707)
            return None

        ## find session
        try:
            pool = self._getHsmPool_(hsm_id)
            self.rwLock.acquire_write()
            found = self._findHSM4Session(pool, sessionId)
            if found is None:
                log.info('could not bind hsm to session %r ' % hsm_id)
            else:
                self._freeHSMSession(pool, sessionId)
        finally:
            self.rwLock.release()
        return True

    def getSecurityModule(self, hsm_id=None, sessionId=None):
        found = None
        if hsm_id is None:
            hsm_id = self.activeOne
        if sessionId is None:
            sessionId = unicode(thread.get_ident())

        if self.config.has_key(hsm_id) == False:
            error = ('[SecurityProvider:getSecurityModule] no config found for '
                     'hsm with id %s ' % (unicode(hsm_id)))
            log.error(error)
            raise HSMException(error, id=707)

        retry = True
        tries = 0
        locked = False

        while retry == True:
            try:
                pool = self._getHsmPool_(hsm_id)
                self.rwLock.acquire_write()
                locked = True
                ## find session
                found = self._findHSM4Session(pool, sessionId)
                if found is not None:
                    ## if session is ok - return
                    self.rwLock.release()
                    locked = False
                    retry = False
                    log.debug("using existing pool session %s" % found)
                    return found.get('obj')
                else:
                    ## create new entry
                    log.debug("getting new Session (%s) "
                              "from pool %s" % (sessionId, pool))
                    found = self._createHSM4Session(pool, sessionId)
                    self.rwLock.release()
                    locked = False
                    if found is None:
                        tries += 1
                        log.warning('try %d: could not bind hsm to session  - '
                                    'going to sleep for  %r' % (tries, 10 * tries))
                        time.sleep(10 * tries)

                        if tries >= self.max_retry:
                            error = ('[SecurityProvider:getSecurityModule] '
                                     'max_retry %d: could not bind hsm to '
                                     'session  - going to sleep for  %r'
                                     % tries, 10 * tries)
                            log.error(error)
                            raise Exception(error)
                        retry = True
                    else:
                        retry = False

            finally:
                if locked == True:
                    self.rwLock.release()

        return found
        #return self.loadSecurityModule(id)


def main():  # pragma: no cover
    # hook for local provider test
    sep = SecurityProvider()
    sep.load_config({})
    sep.createHSMPool('default')
    sep.setupModule('default', {'passwd': 'test123'})

    # runtime catch an hsm for session
    hsm = sep.getSecurityModule()

    passwo = 'password'
    encpass = hsm.encryptPassword(passwo)
    passw = hsm.decryptPassword(encpass)

    zerome(passw)

    hsm2 = sep.getSecurityModule(sessionId='session2')

    passwo = 'password'
    encpass = hsm2.encryptPassword(passwo)
    passw = hsm2.decryptPassword(encpass)

    zerome(passw)

    ## session shutdown
    sep.dropSecurityModule(sessionId='session2')
    sep.dropSecurityModule()

    return True


if __name__ == '__main__':

    main()

#eof###########################################################################




# -*- coding: utf-8 -*-
#
#  2010 - 2014 LSE Leading Security Experts GmbH
#  contact:  http://www.linotp.org
#            http://www.lsexperts.de
#            linotp@lsexperts.de
#
# based on repoze/who/plugins/form.py
#
# Author: Agendaless Consulting <repoze-dev@lists.repoze.org>
#
#
# Copyright: Copyright © 2007 Agendaless Consulting and Contributors
# License:
#   A copyright notice accompanies this license document that identifies
#   the copyright holders.
#  
#   Redistribution and use in source and binary forms, with or without
#   modification, are permitted provided that the following conditions are
#   met:
#  
#   1.  Redistributions in source code must retain the accompanying
#       copyright notice, this list of conditions, and the following
#       disclaimer.
#  
#   2.  Redistributions in binary form must reproduce the accompanying
#       copyright notice, this list of conditions, and the following
#       disclaimer in the documentation and/or other materials provided
#       with the distribution.
#  
#   3.  Names of the copyright holders must not be used to endorse or
#       promote products derived from this software without prior
#       written permission from the copyright holders.
#  
#   4.  If any files are modified, you must cause the modified files to
#       carry prominent notices stating that you changed the files and
#       the date of any change.
#  
#   Disclaimer
#  
#     THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND
#     ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
#     TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
#     PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
#     HOLDERS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
#     EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
#     TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
#     DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
#     ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
#     TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
#     THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
#     SUCH DAMAGE.

import urlparse
import urllib
import cgi

from paste.httpheaders import CONTENT_LENGTH
from paste.httpheaders import CONTENT_TYPE
from paste.httpheaders import LOCATION
from paste.httpexceptions import HTTPFound
from paste.httpexceptions import HTTPUnauthorized

from paste.request import parse_dict_querystring
from paste.request import parse_formvars
from paste.request import construct_url

from paste.response import header_value

from zope.interface import implements

from repoze.who.config import _resolve
from repoze.who.interfaces import IChallenger
from repoze.who.interfaces import IIdentifier

_DEFAULT_FORM = """
<html>
<head>
  <title>Log In</title>
</head>
<body>
  <div>
     <b>Log In</b>
  </div>
  <br/>
  <form method="POST" action="?__do_login=true">
    <table border="0">
    <tr>
      <td>User Name</td>
      <td><input type="text" name="login"></input></td>
    </tr>
    <tr>
      <td>Password</td>
      <td><input type="password" name="password"></input></td>
    </tr>
    <tr>
      <td></td>
      <td><input type="submit" name="submit" value="Log In"/></td>
    </tr>
    </table>
  </form>
  <pre>
  </pre>
</body>
</html>
"""

class FormPluginBase(object):
    def _get_rememberer(self, environ):
        rememberer = environ['repoze.who.plugins'][self.rememberer_name]
        return rememberer

    # IIdentifier
    def remember(self, environ, identity):
        rememberer = self._get_rememberer(environ)
        return rememberer.remember(environ, identity)

    # IIdentifier
    def forget(self, environ, identity):
        rememberer = self._get_rememberer(environ)
        return rememberer.forget(environ, identity)

    def __repr__(self):
        return '<%s %s>' % (self.__class__.__name__,
                            id(self))  #pragma NO COVERAGE

class FormPlugin(FormPluginBase):

    implements(IChallenger, IIdentifier)

    def __init__(self, login_form_qs, rememberer_name, formbody=None,
                 formcallable=None):
        self.login_form_qs = login_form_qs
        # rememberer_name is the name of another configured plugin which
        # implements IIdentifier, to handle remember and forget duties
        # (ala a cookie plugin or a session plugin)
        self.rememberer_name = rememberer_name
        self.formbody = formbody
        self.formcallable = formcallable

    # IIdentifier
    def identify(self, environ):
        query = parse_dict_querystring(environ)
        # If the extractor finds a special query string on any request,
        # it will attempt to find the values in the input body.
        if query.get(self.login_form_qs):
            form = parse_formvars(environ)
            from StringIO import StringIO
            # XXX we need to replace wsgi.input because we've read it
            # this smells funny
            environ['wsgi.input'] = StringIO()
            form.update(query)
            try:
                login = form['login']
                password = form['password']
                realm = form['realm']
            except KeyError:
                return None
            del query[self.login_form_qs]
            environ['QUERY_STRING'] = urllib.urlencode(query)
            environ['repoze.who.application'] = HTTPFound(
                                                    construct_url(environ))
            credentials = {'login':login, 'password':password, 'realm':realm}
            max_age = form.get('max_age', None)
            if max_age is not None:
                credentials['max_age'] = max_age
            return credentials

        return None

    # IChallenger
    def challenge(self, environ, status, app_headers, forget_headers):
        if app_headers:
            location = LOCATION(app_headers)
            if location:
                headers = list(app_headers) + list(forget_headers)
                return HTTPFound(headers=headers)

        form = self.formbody or _DEFAULT_FORM
        if self.formcallable is not None:
            form = self.formcallable(environ)
        def auth_form(environ, start_response):
            content_length = CONTENT_LENGTH.tuples(str(len(form)))
            content_type = CONTENT_TYPE.tuples('text/html')
            headers = content_length + content_type + forget_headers
            start_response('200 OK', headers)
            return [form]

        return auth_form

class RedirectingFormPlugin(FormPluginBase):

    implements(IChallenger, IIdentifier)

    def __init__(self, login_form_url, login_handler_path, logout_handler_path,
                 rememberer_name, reason_param='reason'):
        self.login_form_url = login_form_url
        self.login_handler_path = login_handler_path
        self.logout_handler_path = logout_handler_path
        # rememberer_name is the name of another configured plugin which
        # implements IIdentifier, to handle remember and forget duties
        # (ala a cookie plugin or a session plugin)
        self.rememberer_name = rememberer_name
        self.reason_param = reason_param

    # IIdentifier
    def identify(self, environ):
        path_info = environ['PATH_INFO']
        query = parse_dict_querystring(environ)

        if path_info == self.logout_handler_path:
            # we've been asked to perform a logout
            form = parse_formvars(environ)
            form.update(query)
            referer = environ.get('HTTP_REFERER', '/')
            environ['repoze.who.application'] = HTTPUnauthorized()
            return None

        elif path_info == self.login_handler_path:
            # we've been asked to perform a login
            form = parse_formvars(environ)
            form.update(query)
            try:
                max_age = form.get('max_age', None)
                credentials = {
                    'login':form['login'],
                    'password':form['password'],
                    'realm':form['realm'],
                    }
            except KeyError:
                credentials = None

            if credentials is not None:
                max_age = form.get('max_age', None)
                if max_age is not None:
                    credentials['max_age'] = max_age

            referer = environ.get('HTTP_REFERER', '/')
            environ['repoze.who.application'] = HTTPFound(referer)
            return credentials

    # IChallenger
    def challenge(self, environ, status, app_headers, forget_headers):
        reason = header_value(app_headers, 'X-Authorization-Failure-Reason')
        url_parts = list(urlparse.urlparse(self.login_form_url))
        query = url_parts[4]
        query_elements = cgi.parse_qs(query)
        if reason:
            query_elements[self.reason_param] = reason
        url_parts[4] = urllib.urlencode(query_elements, doseq=True)
        login_form_url = urlparse.urlunparse(url_parts)
        headers = [ ('Location', login_form_url) ]
        cookies = [(h, v) for (h, v) in app_headers if h.lower() == 'set-cookie']
        headers = headers + forget_headers + cookies
        return HTTPFound(headers=headers)

def make_plugin(login_form_qs='__do_login',
                rememberer_name=None,
                form=None,
                formcallable=None,
               ):
    if rememberer_name is None:
        raise ValueError(
            'must include rememberer key (name of another IIdentifier plugin)')
    if form is not None:
        form = open(form).read()
    if isinstance(formcallable, str):
        formcallable = _resolve(formcallable)
    plugin = FormPlugin(login_form_qs, rememberer_name, form, formcallable)
    return plugin

def make_redirecting_plugin(login_form_url=None,
                            login_handler_path='/login_handler',
                            logout_handler_path='/logout_handler',
                            rememberer_name=None):
    if login_form_url is None:
        raise ValueError(
            'must include login_form_url in configuration')
    if login_handler_path is None:
        raise ValueError(
            'login_handler_path must not be None')
    if logout_handler_path is None:
        raise ValueError(
            'logout_handler_path must not be None')
    if rememberer_name is None:
        raise ValueError(
            'must include rememberer key (name of another IIdentifier plugin)')
    plugin = RedirectingFormPlugin(login_form_url,
                                   login_handler_path,
                                   logout_handler_path,
                                   rememberer_name)
    return plugin




# -*- coding: utf-8 -*-
#
#  privacyIDEA is a fork of LinOTP
#  May 08, 2014 Cornelius Kölbel
#  License:  AGPLv3
#  contact:  http://www.privacyidea.org
#
#  Copyright (C) 2010 - 2014 LSE Leading Security Experts GmbH
#  License:  AGPLv3
#  contact:  http://www.linotp.org
#            http://www.lsexperts.de
#            linotp@lsexperts.de
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
'''
This file containes the Config object of the request.
'''



import logging
import time
import os
import copy

from pylons import tmpl_context as c
from privacyidea.config import environment as env
from privacyidea.lib.log import log_with
from privacyidea.lib.error import ConfigAdminError

from privacyidea.model import Config
from privacyidea.model.meta import Session

from privacyidea.lib.crypto import encryptPassword
from privacyidea.lib.crypto import decryptPassword

from datetime import datetime



log = logging.getLogger(__name__)

ENCODING = 'utf-8'


###############################################################################
##     public interface
###############################################################################

@log_with(log)
def init_privacyIDEA_config():
    '''
    return the privacyideaConfig class, which is integrated
    in the local thread context

    :return: thread local privacyIDEAConfig
    :rtype:  privacyIDEAConfig Class
    '''
    ret = get_privacyIDEA_config()
    return ret

@log_with(log)
def get_privacyIDEA_config():
    '''
    return the thread local dict with all entries

    :return: local config dict
    :rtype: dict
    '''

    ret = {}
    try:
        if False == hasattr(c, 'privacyideaConfig'):
            c.privacyideaConfig = privacyIDEAConfig()

        ty = type(c.privacyideaConfig).__name__
        if ty != 'privacyIDEAConfig':
            try:
                c.privacyideaConfig = privacyIDEAConfig()
            except Exception as e:
                log.error("privacyIDEA Definition Error")
                raise Exception(e)
        ret = c.privacyideaConfig

        if ret.delay == True:
            if hasattr(c, 'hsm') == True and isinstance(c.hsm, dict):
                hsm = c.hsm.get('obj')
                if hsm is not None and hsm.isReady() == True:
                    ret = privacyIDEAConfig()
                    c.privacyideaConfig = ret

    except Exception as e:
        log.debug("Bad Hack: privacyIDEAConfig called out of controller context")
        ret = privacyIDEAConfig()

        if ret.delay == True:
            if hasattr(c, 'hsm') == True and isinstance(c.hsm, dict):
                hsm = c.hsm.get('obj')
                if hsm is not None and hsm.isReady() == True:
                    ret = privacyIDEAConfig()

    return ret

###############################################################################
##     implementation class
###############################################################################
class privacyIDEAConfig(dict):
    '''
    this class should be a request singleton.

     In case of a change, it must cover the different aspects like
    - env config entry   and
    - app_globals
    and finally
    - sync this to disc


    '''
    @log_with(log)
    def __init__(self, *args, **kw):
        self.parent = super(privacyIDEAConfig, self)
        self.parent.__init__(*args, **kw)

        self.delay = False
        self.realms = None
        self.glo = getGlobalObject()
        conf = self.glo.getConfig()

        do_reload = False

        # do the bootstrap if no entry in the app_globals
        if len(conf.keys()) == 0:
            do_reload = True

        if self.glo.isConfigComplet() == False:
            do_reload = True
            self.delay = True

        if 'privacyidea.enableReplication' in conf:
            val = conf.get('privacyidea.enableReplication')
            if val.lower() == 'true':

                ## look for the timestamp when config was created
                e_conf_date = conf.get('privacyidea.Config')

                ## in case of replication, we always have to look if the
                ## config data in the database changed
                db_conf_date = _retrieveConfigDB('privacyidea.Config')

                if str(db_conf_date) != str(e_conf_date):
                    do_reload = True

        if do_reload == True:
            ## in case there is no entry in the dbconf or
            ## the config file is newer, we write the config back to the db
            entries = conf.keys()
            for entry in entries:
                del conf[entry]

            writeback = False
            ## get all conf entries from the config file
            fileconf = _getConfigFromEnv()

            ##  get all configs from the DB
            (dbconf, delay) = _retrieveAllConfigDB()
            self.glo.setConfigIncomplete(not delay)

            ## we only merge the config file once as a removed entry
            ##  might reappear otherwise
            if dbconf.has_key('privacyidea.Config') == False:
                conf.update(fileconf)
                writeback = True
            ##
            ##else:
            ##    modCFFileDatum = fileconf.get('privacyidea.Config')
            ##    dbTimeStr = dbconf.get('privacyidea.Config')
            ##    dbTimeStr = dbTimeStr.split('.')[0]
            ##    modDBFileDatum =
            ##           datetime.strptime(dbTimeStr,'%Y-%m-%d %H:%M:%S')
            ##    # if configFile timestamp is newer than last update:
            ##    #             reincorporate conf
            ##    #if modCFFileDatum > modDBFileDatum:
            ##    #    conf.update(fileconf)
            ##    #    writeback = True
            ##

            conf.update(dbconf)
            ## chck, if there is a selfTest in the DB and delete it
            if dbconf.has_key('privacyidea.selfTest'):
                _removeConfigDB('privacyidea.selfTest')
                _storeConfigDB('privacyidea.Config', datetime.now())

            ## the only thing we take from the fileconf is the selftest
            if fileconf.has_key('privacyidea.selfTest'):
                conf['privacyidea.selfTest'] = 'True'

            if writeback == True:
                for con in conf:
                    if con != 'privacyidea.selfTest':
                        _storeConfigDB(con, conf.get(con))
                _storeConfigDB('privacyidea.Config', datetime.now())

            self.glo.setConfig(conf, replace=True)

        self.parent.update(conf)
        return

    def setRealms(self, realmDict):
        self.realms = realmDict
        return

    def getRealms(self):
        return self.realms

    @log_with(log)
    def addEntry(self, key, val, typ=None, des=None):
        '''
        small wrapper, as the assignement opperator has only one value argument

        :param key: key of the dict
        :type  key: string
        :param val: any value, which is put in the dict
        :type  val: any type
        :param typ: used in the database to control if the data is encrypted
        :type  typ: None,string,password
        :param des: literal, which describes the data
        :type  des: string
        '''
        if key.startswith('privacyidea.') == False:
            key = 'privacyidea.' + key

        if type(val) in [str, unicode] and "%(here)s" in val:
            val = _expandHere(val)

        res = self.__setitem__(key, val, typ, des)
        return res

    @log_with(log)
    def __setitem__(self, key, val, typ=None, des=None):
        '''
        implemtation of the assignement operator == internal function

        :param key: key of the dict
        :type  key: string
        :param val: any value, which is put in the dict
        :type  val: any type
        :param typ: used in the database to control if the data is encrypted
        :type  typ: None,string,password
        :param des: literal, which describes the data
        :type  des: string
        '''

        if typ == 'password':

            ## in case we have a password type, we have to put
            ##- in the config only the encrypted pass and
            ##- add the config encprivacyidea.* with the clear password

            res = self.parent.__setitem__(key, encryptPassword(val))
            res = self.parent.__setitem__('enc' + key, val)
            self.glo.setConfig({key :encryptPassword(val)})
            self.glo.setConfig({'enc' + key : val})

        else:
            ## update this config and sync with global dict and db
            nVal = _expandHere(val)
            res = self.parent.__setitem__(key, nVal)
            self.glo.setConfig({key:nVal})

        _storeConfigDB(key, val, typ, des)
        _storeConfigDB('privacyidea.Config', datetime.now())
        return res

    def get(self, key, default=None):
        '''
            check for a key in the privacyidea config

            remark: the config entries all start with privacyidea.
            if a key is not found, we do a check if there is
            a privacyidea. prefix set in the key and potetialy prepend it

            :param key: search value
            :type  key: string
            :param default: default value, which is returned,
                            if the value is not found
            :type  default: any type

            :return: value or None
            :rtype:  any type
        '''
        if (self.parent.has_key(key) == False
                and key.startswith('privacyidea.') == False):
            key = 'privacyidea.' + key
        res = self.parent.get(key) or default
        return res

    def has_key(self, key):
        res = self.parent.has_key(key)
        if res == False and key.startswith('privacyidea.') == False:
            key = 'privacyidea.' + key

        res = self.parent.has_key(key)

        if res == False and key.startswith('encprivacyidea.') == False:
            key = 'encprivacyidea.' + key

        res = self.parent.has_key(key)

        return res

    @log_with(log)
    def __delitem__(self, key):
        '''
        remove an item from the config

        :param key: the name of the ocnfig entry
        :type  key: string

        :return : return the std value like the std dict does, whatever this is
        :rtype  : any value a dict update will return
        '''
        Key = key
        encKey = None
        if self.parent.has_key(key):
            Key = key
        elif self.parent.has_key('privacyidea.' + key):
            Key = 'privacyidea.' + key

        if self.parent.has_key('encprivacyidea.' + key):
            encKey = 'encprivacyidea.' + key
        elif self.parent.has_key('enc' + key):
            encKey = 'enc' + key

        res = self.parent.__delitem__(Key)
        ## sync with global dict
        self.glo.delConfig(Key)

        ## do we have an decrypted in local or global dict??
        if encKey is not None:
            res = self.parent.__delitem__(encKey)
            ## sync with global dict
            self.glo.delConfig(encKey)

        ## sync with db
        if key.startswith('privacyidea.'):
            Key = key
        else:
            Key = 'privacyidea.' + key

        _removeConfigDB(Key)
        _storeConfigDB('privacyidea.Config', datetime.now())
        return res

    @log_with(log)
    def update(self, dic):
        '''
        update the config dict with multiple items in a dict

        :param dic: dictionary of multiple items
        :type  dic: dict

        :return : return the std value like the std dict does, whatever this is
        :rtype  : any value a dict update will return
        '''
        res = self.parent.update(dic)
        ## sync the lobal dict
        self.glo.setConfig(dic)
        ## sync to disc
        for key in dic:
            if key != 'privacyidea.Config':
                _storeConfigDB(key, dic.get(key))

        _storeConfigDB('privacyidea.Config', datetime.now())
        return res


###############################################################################
##  helper class from here
###############################################################################
def getGlobalObject():
    glo = None

    try:
        if env.config.has_key('pylons.app_globals'):
            glo = env.config['pylons.app_globals']
        elif env.config.has_key('pylons.g'):
            glo = env.config['pylons.g']
    except:
        glo = None
    return glo

def _getConfigReadLock():
    glo = getGlobalObject()
    rcount = glo.setConfigReadLock()
    log.debug(" --------------------------------------- Read Lock %s" % rcount)

def _getConfigWriteLock():
    glo = getGlobalObject()
    rcount = glo.setConfigWriteLock()
    log.debug(" ------------------- ------------------ Write Lock %s" % rcount)

def _releaseConfigLock():
    glo = getGlobalObject()
    rcount = glo.releaseConfigLock()
    log.debug(" ------------------------------------ release Lock %s" % rcount)


@log_with(log)
def _expandHere(value):
    Value = unicode(value)
    if env.config.has_key("privacyidea.root"):
        root = env.config["privacyidea.root"]
        Value = Value.replace("%(here)s", root)
    return Value

@log_with(log)
def _getConfigFromEnv():
    privacyideaConfig = {}

    try:
        _getConfigReadLock()
        for entry in env.config:
            ## we check for the modification time of the config file
            if entry == '__file__':
                fname = env.config.get('__file__')
                mTime = time.localtime(os.path.getmtime(fname))
                modTime = datetime(*mTime[:6])
                privacyideaConfig['privacyidea.Config'] = modTime

            if entry.startswith("privacyidea."):
                privacyideaConfig[entry] = _expandHere(env.config[entry])
            if entry.startswith("encprivacyidea."):
                privacyideaConfig[entry] = env.config[entry]
        _releaseConfigLock()
    except Exception as e:
        log.error('Error while reading Config: %r' % e)
        _releaseConfigLock()
    return privacyideaConfig


# we insert or update the key / value the config DB
@log_with(log)
def _storeConfigDB(key, val, typ=None, desc=None):
    value = val

    if (not key.startswith("privacyidea.")):
        key = "privacyidea." + key

    confEntries = Session.query(Config).filter(Config.Key == unicode(key))
    theConf = None

    if typ is not None and typ == 'password':
        value = encryptPassword(val)
        en = decryptPassword(value)
        if (en != val):
            raise Exception("StoreConfig: Error during encoding password type!")

    ## update
    if confEntries.count() == 1:
        theConf = confEntries[0]
        theConf.Value = unicode(value)
        if (typ is not None):
            theConf.Type = unicode(typ)
        if (desc is not None):
            theConf.Description = unicode(desc)

    ## insert
    elif confEntries.count() == 0:
        theConf = Config(
                        Key=unicode(key),
                        Value=unicode(value),
                        Type=unicode(typ),
                        Description=unicode(desc)
                        )
    if theConf is not None:
        Session.add(theConf)

    return 101

@log_with(log)
def _removeConfigDB(key):
    if (not key.startswith("privacyidea.")):
        if not key.startswith('encprivacyidea.'):
            key = u"privacyidea." + key

    confEntries = Session.query(Config).filter(Config.Key == unicode(key))
    num = confEntries.count()
    if num == 1:
        theConf = confEntries[0]

        try:
            #Session.add(theConf)
            Session.delete(theConf)

        except Exception as e:
            log.error('failed')
            raise ConfigAdminError("remove Config failed for %r: %r"
                                   % (key, e), id=1133)

    return num

@log_with(log)
def _retrieveConfigDB(Key):
    ## prepend "lonotp." if required
    key = Key
    if (not key.startswith("privacyidea.")):
        if (not key.startswith("encprivacyidea.")):
            key = "privacyidea." + Key

    myVal = None
    key = u'' + key
    for theConf in Session.query(Config).filter(Config.Key == key):
        myVal = theConf.Value
        myVal = _expandHere(myVal)
    return myVal

@log_with(log)
def _retrieveAllConfigDB():

    config = {}
    delay = False
    for conf in Session.query(Config).all():
        log.debug("key %r:%r" % (conf.Key, conf.Value))
        key = conf.Key
        if (not key.startswith("privacyidea.")):
            key = "privacyidea." + conf.Key
        nVal = _expandHere(conf.Value)
        config[key] = nVal
        myTyp = conf.Type
        if myTyp is not None:
            if myTyp == 'password':
                if hasattr(c, 'hsm') == True and isinstance(c.hsm, dict):
                    hsm = c.hsm.get('obj')
                    if hsm is not None and hsm.isReady() == True:
                        config['enc' + key] = decryptPassword(conf.Value)
                else:
                    delay = True

    return (config, delay)

########### external interfaces ###############
@log_with(log)
def storeConfig(key, val, typ=None, desc=None):
    conf = get_privacyIDEA_config()
    conf.addEntry(key, val, typ, desc)
    return True

@log_with(log)
def updateConfig(confi):
    '''
    update the server config entries incl. syncing it to disc
    '''
    conf = get_privacyIDEA_config()

    ## remember all key, which should be processed
    p_keys = copy.deepcopy(confi)

    typing = False

    for entry in confi:
        typ = confi.get(entry + ".type", None)
        des = confi.get(entry + ".desc", None)
        ## check if we have a descriptive entry
        if typ is not None or des is not None:
            typing = True
            if typ is not None:
                del p_keys[entry + ".type"]
            if des is not None:
                del p_keys[entry + ".desc"]

    if typing == True:
        ## tuple dict containing the additional info
        t_dict = {}
        for entry in p_keys:
            val = confi.get(entry)
            typ = confi.get(entry + ".type", None)
            des = confi.get(entry + ".desc", None)
            t_dict[entry] = (val, typ or "string", des)

        for key in t_dict:
            (val, typ, desc) = t_dict.get(key)
            if val in [str, unicode] and "%(here)s" in val:
                val = _expandHere(val)
            conf.addEntry(key, val, typ, desc)

    else:
        conf_clean = {}
        for key, val in confi.iteritems():
            if "%(here)s" in val:
                val = _expandHere(val)
            conf_clean[key] = val

        conf.update(conf_clean)

    return True
    
@log_with(log)    
def getFromConfig(key, defVal=None):
    conf = get_privacyIDEA_config()
    value = conf.get(key, defVal)
    return value

@log_with(log)    
def removeFromConfig(key, iCase=False):
    conf = get_privacyIDEA_config()

    if iCase == False:
        if conf.has_key(key):
            del conf[key]
    else:
        ## case insensitive delete
        ##- might have multiple hits
        fConf = []
        for k in conf:
            if (k.lower() == key.lower() or
                  k.lower() == 'privacyidea.' + key.lower()):
                fConf.append(k)

        if len(fConf) > 0:
            for k in fConf:
                if conf.has_key(k) or conf.has_key('privacyidea.' + k):
                    del conf[k]

    return True


#### several config functions to follow
@log_with(log)
def setDefaultMaxFailCount(maxFailCount):
    return storeConfig(u"DefaultMaxFailCount", maxFailCount)

@log_with(log)
def setDefaultSyncWindow(syncWindowSize):
    return storeConfig(u"DefaultSyncWindow", syncWindowSize)

@log_with(log)
def setDefaultCountWindow(countWindowSize):
    return storeConfig(u"DefaultCountWindow", countWindowSize)

@log_with(log)
def setDefaultOtpLen(otpLen):
    return storeConfig(u"DefaultOtpLen", otpLen)

@log_with(log)
def setDefaultResetFailCount(resetFailCount):
    return storeConfig(u"DefaultResetFailCount", resetFailCount)


#eof###########################################################################




# -*- coding: utf-8 -*-
#
#  product:  privacyIDEA is a fork of LinOTP
#  module:   privacyidea resolver library
#
#  May, 08 2014 Cornelius Kölbel
#  http://www.privacyidea.org
#
# 2014-10-03 fix getUsername function
#            Cornelius Kölbel <cornelius@privcyidea.org>
#
#  product:  LinOTP2
#  module:   useridresolver
#  tool:     PasswdIdResolver
#  edition:  Comunity Edition
#
#  Copyright (C) 2010 - 2014 LSE Leading Security Experts GmbH
#  License:  AGPLv3
#  contact:  http://www.linotp.org
#            http://www.lsexperts.de
#            linotp@lsexperts.de
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
'''
  Description:  This file is part of the privacyidea service
                This module implements the communication interface
                for resolvin user info to the /etc/passwd user base

  Dependencies: -
'''

import re
import logging
import crypt


from UserIdResolver import UserIdResolver
from UserIdResolver import getResolverClass


log = logging.getLogger(__name__)
ENCODING = "utf-8"


def tokenise(r):
    def _(s):
        ret = None
        st = s.strip()
        m = re.match("^" + r, st)
        if m:
            ret = (st[:m.end()].strip(), st[m.end():].strip())
        return ret
    return _


class IdResolver (UserIdResolver):

    fields = {"username": 1, "userid": 1,
              "description": 0,
              "phone": 0, "mobile": 0, "email": 0,
              "givenname": 0, "surname": 0, "gender": 0
              }

    searchFields = {"username": "text",
                    "userid": "numeric",
                    "description": "text",
                    "email": "text"
                    }

    sF = {"username": 0,
          "cryptpass": 1,
          "userid": 2,
          "description": 4,
          "email": 4,
          }

    @classmethod
    def setup(cls, config=None, cache_dir=None):
        '''
        this setup hook is triggered, when the server
        starts to serve the first request

        :param config: the privacyidea config
        :type  config: the privacyidea config dict
        '''
        log.info("Setting up the PasswdResolver")
        return

    def __init__(self):
        """
        simple constructor
        """
        self.name = "etc-passwd"
        self.fileName = ""

        self.name = "P"
        self.nameDict = {}
        self.descDict = {}
        self.reversDict = {}
        self.passDict = {}
        self.officePhoneDict = {}
        self.homePhoneDict = {}
        self.surnameDict = {}
        self.givennameDict = {}
        self.emailDict = {}

    def close(self):
        """
        request hook - to close down resolver object
        """
        return

    def loadFile(self):

        """
          init loads the /etc/passwd
            user and uid as a dict for /
            user loginname lookup
        """

        if (self.fileName == ""):
            self.fileName = "/etc/passwd"

        log.info('loading users from file %s' % (self.fileName))

        fileHandle = open(self.fileName, "r")

        line = fileHandle.readline()

        ID = self.sF["userid"]
        NAME = self.sF["username"]
        PASS = self.sF["cryptpass"]
        DESCRIPTION = self.sF["description"]

        while line:
            line = line.strip()
            if len(line) == 0:
                continue

            fields = line.split(":", 7)
            self.nameDict["%s" % fields[NAME]] = fields[ID]

            # for speed reason - build a revers lookup
            self.reversDict[fields[ID]] = "%s" % fields[NAME]

            # for full info store the line
            self.descDict[fields[ID]] = fields

            # store the crypted password
            self.passDict[fields[ID]] = fields[PASS]

            # store surname, givenname and phones
            descriptions = fields[DESCRIPTION].split(",")
            name = descriptions[0]
            names = name.split(' ', 1)
            self.givennameDict[fields[ID]] = names[0]
            self.surnameDict[fields[ID]] = ""
            self.officePhoneDict[fields[ID]] = ""
            self.homePhoneDict[fields[ID]] = ""
            self.emailDict[fields[ID]] = ""
            if len(names) >= 2:
                self.surnameDict[fields[ID]] = names[1]
            if len(descriptions) >= 4:
                self.officePhoneDict[fields[ID]] = descriptions[2]
                self.homePhoneDict[fields[ID]] = descriptions[3]
            if len(descriptions) >= 5:
                for field in descriptions[4:]:
                    # very basic e-mail regex
                    email_match = re.search('.+@.+\..+', field)
                    if email_match:
                        self.emailDict[fields[ID]] = email_match.group(0)

            """ print ">>" + key[0] + "<< " + key[2] """
            line = fileHandle.readline()
            
        fileHandle.close()

    def checkPass(self, uid, password):
        """
        This function checks the password for a given uid.
        - returns true in case of success
        -         false if password does not match

        We do not support shadow passwords at the moment. so the seconds column
        of the passwd file needs to contain the crypted password
        """
        log.info("checking password for user uid %s" % uid)
        cryptedpasswd = self.passDict[uid]
        log.debug("We found the crypted pass %s for uid %s"
                  % (cryptedpasswd, uid))
        if cryptedpasswd:
            if cryptedpasswd == 'x' or cryptedpasswd == '*':
                err = "Sorry, currently no support for shadow passwords"
                log.error("%s" % err)
                raise NotImplementedError(err)
            cp = crypt.crypt(password, cryptedpasswd)
            log.debug("crypted pass is %s" % cp)
            if crypt.crypt(password, cryptedpasswd) == cryptedpasswd:
                log.info("successfully authenticated user uid %s" % uid)
                return True
            else:
                log.warning("user uid %s failed to authenticate" % uid)
                return False
        else:
            log.warning("Failed to verify password. No crypted password "
                        "found in file")
            return False

    def getUserInfo(self, userId, no_passwd=False):
        """
        get some info about the user
        as we only have the loginId, we have to traverse the dict for the value

        :param userId: the to be searched user
        :param no_passwd: retrun no password
        :return: dict of user info
        """
        ret = {}

        if userId in self.reversDict:
            fields = self.descDict.get(userId)

            for key in self.sF:
                if no_passwd and key == "cryptpass":
                    continue
                index = self.sF[key]
                ret[key] = fields[index]

            ret['givenname'] = self.givennameDict.get(userId)
            ret['surname'] = self.surnameDict.get(userId)
            ret['phone'] = self.homePhoneDict.get(userId)
            ret['mobile'] = self.officePhoneDict.get(userId)
            ret['email'] = self.emailDict.get(userId)

        return ret

    def getUsername(self, userId):
        '''
        Returns the username/loginname for a given userid
        :param userid: The userid in this resolver
        :type userid: string
        :return: username
        :rtype: string
        '''
        fields = self.descDict.get(userId)
        index = self.sF["username"]
        return fields[index]

    def getUserId(self, LoginName):
        """
        search the user id from the login name

        :param LoginName: the login of the user
        :return: the userId
        """
        # We need the encoding, to be also able to read usernames
        # with Umlauts from files.

        if type(LoginName) == unicode:
            LoginName = LoginName.encode(ENCODING)

        if LoginName in self.nameDict.keys():
            return self.nameDict[LoginName]
        else:
            return ""

    def getSearchFields(self, searchDict=None):
        """
        show, which search fields this userIdResolver supports

        TODO: implementation is not completed

        :param searchDict: fields, which should be queried
        :return: dict of all searchFields
        """
        if searchDict is not None:
            for search in searchDict:
                pattern = searchDict[search]

                log.debug("searching for %s:%s",
                          search, pattern)

        return self.searchFields

    def getUserList(self, searchDict):
        """
        get a list of all users matching the search criteria of the searchdict

        :param searchDict: dict of search expressions
        """
        ret = []

        #  first check if the searches are in the searchDict
        for l in self.descDict:
            line = self.descDict[l]
            ok = True

            for search in searchDict:

                if search not in self.searchFields:
                    ok = False
                    break

                pattern = searchDict[search]

                log.debug("searching for %s:%s", search, pattern)

                if search == "username":
                    ok = self.checkUserName(line, pattern)
                elif search == "userid":
                    ok = self.checkUserId(line, pattern)
                elif search == "description":
                    ok = self.checkDescription(line, pattern)
                elif search == "email":
                    ok = self.checkEmail(line, pattern)

                if ok is not True:
                    break

            if ok is True:
                uid = line[self.sF["userid"]]
                info = self.getUserInfo(uid, no_passwd=True)
                ret.append(info)

        return ret

    def checkUserName(self, line, pattern):
        """
        check for user name
        """

        username = line[self.sF["username"]]
        ret = self.stringMatch(username, pattern)
        return ret

    def checkDescription(self, line, pattern):
        description = line[self.sF["description"]]
        ret = self.stringMatch(description, pattern)
        return ret

    def checkEmail(self, line, pattern):
        email = line[self.sF["email"]]
        ret = self.stringMatch(email, pattern)
        return ret

    def stringMatch(self, cString, cPattern):
        ret = False
        e = s = ""

        if type(cString) == unicode:
            cString = cString.encode(ENCODING)

        if type(cPattern) == unicode:
            cPattern = cPattern.encode(ENCODING)

        string = cString.lower()
        pattern = cPattern.lower()

        if pattern.startswith("*"):
            e = "e"
            pattern = pattern[1:]

        if pattern.endswith("*"):
            s = "s"
            pattern = pattern[:-1]

        if (e == "e" and s == "s"):
            if string.find(pattern) != -1:
                return True
        elif (e == "e"):
            if string.endswith(pattern):
                return True
        elif (s == "s"):
            if string.startswith(pattern):
                return True
        else:
            if string == pattern:
                return True

        return ret

    def checkUserId(self, line, pattern):
        """
        check for the userId
        """
        ret = False

        try:
            cUserId = int(line[self.sF["userid"]])
        except:
            return ret

        (op, val) = tokenise(">=|<=|>|<|=|between")(pattern)

        if op == "between":
            (lVal, hVal) = val.split(",", 2)
            try:
                ilVal = int(lVal.strip())
                ihVal = int(hVal.strip())
                if ihVal < ilVal:
                    v = ihVal
                    ihVal = ilVal
                    ilVal = v
            except:
                return ret

            if (cUserId <= ihVal and cUserId >= ilVal):
                ret = True
        else:
            try:
                ival = int(val)
            except:
                return ret

            if op == "=":
                if (cUserId == ival):
                    ret = True

            elif op == ">":
                if (cUserId > ival):
                    ret = True

            elif op == ">=":
                if (cUserId >= ival):
                    ret = True

            elif op == "<":
                if (cUserId < ival):
                    ret = True

            elif op == "<=":
                if (cUserId < ival):
                    ret = True

        return ret

#############################################################
# server info methods
#############################################################
    def getResolverId(self):
        """ getResolverId(LoginName)
            - returns the resolver identifier string
            - empty string if not exist
        """
        return self.fileName

    @classmethod
    def getResolverClassType(cls):
        return 'passwdresolver'

    def getResolverType(self):
        return IdResolver.getResolverClassType()

    @classmethod
    def getResolverClassDescriptor(cls):
        '''
        return the descriptor of the resolver, which is
        - the class name and
        - the config description

        :return: resolver description dict
        :rtype:  dict
        '''
        descriptor = {}
        typ = cls.getResolverClassType()
        descriptor['clazz'] = "useridresolver.PasswdIdResolver.IdResolver"
        descriptor['config'] = {'fileName': 'string'}
        return {typ: descriptor}

    def getResolverDescriptor(self):
        return IdResolver.getResolverClassDescriptor()

    def getConfigEntry(self, config, key, conf, required=True):
        ckey = key
        cval = ""
        if conf != "" or None:
            ckey = ckey + "." + conf
            if ckey in config:
                cval = config[ckey]
        if cval == "":
            if key in config:
                cval = config[key]
        if cval == "" and required is True:
            raise Exception("missing config entry: " + key)
        return cval

    def loadConfig(self, config, conf):
        """ loadConfig(configDict)
            The UserIdResolver could be configured
            from the pylons app config - here
            this could be the passwd file ,
            whether it is /etc/passwd or /etc/shadow
        """
        self.fileName = self.getConfigEntry(config,
                                            'privacyidea.passwdresolver.'
                                            'fileName',
                                            conf)
        self.loadFile()

        return self

if __name__ == "__main__":

    print " PasswdIdResolver - IdResolver class test "

    y = getResolverClass("PasswdIdResolver", "IdResolver")()

    y.loadConfig({'privacyidea.passwdresolver.fileName': '/etc/passwd'}, "")
    x = getResolverClass("PasswdIdResolver", "IdResolver")()
    x.loadConfig({'privacyidea.passwdresolver.fileName': '/etc/meinpass'}, "")

    print "======/etc/meinpass=========="
    print x.getUserList({'username': '*', "userid": ">= 1000"})
    print "======/etc/passwd=========="
    print y.getUserList({'username': '*', "userid": ">= 1000"})
    print "================"

    user = "koelbel"
    loginId = y.getUserId(user)

    print " %s -  %s" % (user, loginId)
    print " reId - " + y.getResolverId()

    ret = y.getUserInfo(loginId)

    print "result %r" % ret

    ret = y.getSearchFields()
    search = {"userid": " between 1000, 1005",
              }

    ret = y.getUserList(search)

    print ret
